#!/usr/bin/env python3
#

''' Implementation of directories (Dir) and their entries (FileDirent, etc).
'''

from enum import IntEnum, IntFlag
from functools import partial
import grp
import os
import os.path
import pwd
import stat
import sys
import time
from uuid import UUID, uuid4
from cs.binary import PacketField, BSUInt, BSString, BSData
from cs.buffer import CornuCopyBuffer
from cs.excutils import logexc
from cs.logutils import debug, error, warning, info
from cs.pfx import Pfx
from cs.lex import texthexify
from cs.py.func import prop
from cs.py.stack import stack_dump
from cs.queues import MultiOpenMixin
from cs.threads import locked
from . import totext, PATHSEP, defaults, RLock
from .block import Block, _Block, BlockRecord
from .blockify import top_block_for, blockify
from .file import RWBlockFile
from .meta import Meta, DEFAULT_DIR_ACL, DEFAULT_FILE_ACL
from .paths import path_split, DirLike, FileLike
from .transcribe import Transcriber, parse as parse_transcription, \
                        register as register_transcriber

uid_nobody = -1
gid_nogroup = -1

# Directories (Dir, a subclass of dict) and directory entries (_Dirent).

class DirentType(IntEnum):
  ''' Dirent type codes.
  '''
  INVALID = -1
  FILE = 0
  DIR = 1
  SYMBOLIC = 2
  INDIRECT = 3

class DirentFlags(IntFlag):
  ''' Flag values for the Dirent binary encoding.
  '''
  HASMETA = 0x01        # has metadata
  HASNAME = 0x02        # has a name
  NOBLOCK = 0x04        # has no Block reference
  HASUUID = 0x08        # has a UUID
  HASPREVDIRENT = 0x10  # has reference to serialised previous Dirent
  EXTENDED = 0x20       # extended BSData field

class DirentRecord(PacketField):
  ''' PacketField subclass to parsing and transcribing Dirents in binary form.

      The serialisation format is:

            BSUint(type)
            BSUint(flags)
            [BSString(name)]
            [BSString(str(meta))]
            [uuid:16]
            blockref
            [blockref(pref_dirent)]
            [BSData(extended_data)]

      Note that all additional future implementation detail needs
      to go in the metadata or the optional extended_data.
  '''

  @classmethod
  def value_from_buffer(cls, bfr):
    ''' Unserialise a serialised Dirent.
    '''
    type_ = BSUInt.value_from_buffer(bfr)
    flags = DirentFlags(BSUInt.value_from_buffer(bfr))
    if flags & DirentFlags.HASNAME:
      flags ^= DirentFlags.HASNAME
      name = BSString.value_from_buffer(bfr)
    else:
      name = ""
    if flags & DirentFlags.HASMETA:
      flags ^= DirentFlags.HASMETA
      metatext = BSString.value_from_buffer(bfr)
    else:
      metatext = None
    uu = None
    if flags & DirentFlags.HASUUID:
      flags ^= DirentFlags.HASUUID
      uu = UUID(bytes=bfr.take(16))
    if flags & DirentFlags.NOBLOCK:
      flags ^= DirentFlags.NOBLOCK
      block = None
    else:
      block = BlockRecord.value_from_buffer(bfr)
    if flags & DirentFlags.HASPREVDIRENT:
      flags ^= DirentFlags.HASPREVDIRENT
      prev_dirent_blockref = BlockRecord.value_from_buffer(bfr)
    else:
      prev_dirent_blockref = None
    if flags & DirentFlags.EXTENDED:
      flags ^= DirentFlags.EXTENDED
      extended_data = BSData.value_from_buffer(bfr)
    else:
      extended_data = None
    if flags:
      warning(
          "%s.value_from_buffer: unexpected extra flags: 0x%02x",
          cls.__name__, flags)
    E = _Dirent.from_components(type_, name, meta=metatext, uuid=uu, block=block)
    E._prev_dirent_blockref = prev_dirent_blockref
    E.ingest_extended_data(extended_data)
    return E

  @staticmethod
  def transcribe_value(E):
    ''' Serialise to binary format.
    '''
    flags = 0
    type_ = E.type
    if E.name:
      flags |= DirentFlags.HASNAME
    meta = None if E.isindirect else E.meta
    if meta:
      flags |= DirentFlags.HASMETA
    if E.uuid:
      flags |= DirentFlags.HASUUID
    block = None if type_ is DirentType.INDIRECT else E.block
    if block is None:
      flags |= DirentFlags.NOBLOCK
    if E._prev_dirent_blockref is not None:
      flags |= DirentFlags.HASPREVDIRENT
    extended_data = E.get_extended_data()
    if extended_data:
      flags |= DirentFlags.EXTENDED
    yield BSUInt.transcribe_value(type_)
    yield BSUInt.transcribe_value(flags)
    if flags & DirentFlags.HASNAME:
      yield BSString.transcribe_value(E.name)
    if flags & DirentFlags.HASMETA:
      yield BSString.transcribe_value(meta.textencode())
    if flags & DirentFlags.HASUUID:
      bs = E.uuid.bytes
      if len(bs) != 16:
        raise RuntimeError("len(E.uuid.bytes) != 16: %r" % (bs,))
      yield bs
    if not flags & DirentFlags.NOBLOCK:
      yield BlockRecord.transcribe_value(block)
    if flags & DirentFlags.HASPREVDIRENT:
      assert isinstance(E._prev_dirent_blockref, _Block)
      yield BlockRecord.transcribe_value(E._prev_dirent_blockref)
    if flags & DirentFlags.EXTENDED:
      yield extended_data

class _Dirent(Transcriber):
  ''' Incomplete base class for Dirent objects.
  '''

  transcribe_prefix = 'DIRENT'

  def __init__(
      self,
      type_, name,
      *,
      meta=None,
      uuid=None,
      parent=None,
      prevblock=None,
      block=None,
      **kw
  ):
    ''' Initialise a _Dirent.

        Parameters:
        * `type_`: the DirentType enum
        * `name`: the Dirent's name
        * `meta`: optional metadata
        * `uuid`: optional identifying UUID;
          *note*: for IndirectBlocks this is a reference to another Dirent's
          UUID.
        * `parent`: optional parent Dirent
        * `prevblock`: optional Block whose contents are the binary
          transcription of this Dirent's previous state - another
          Dirent
    '''
    with Pfx("_Dirent(type_=%s,name=%r,...)", type_, name):
      if not isinstance(type_, int):
        raise TypeError("type_ is not an int: <%s>%r" % (type(type_), type_))
      if name is not None and not isinstance(name, str):
        raise TypeError("name is neither None nor str: <%s>%r" % (type(name), name))
      if kw:
        error("unsupported keyword arguments: %r", kw)
      if block is not None:
        raise ValueError("block is not None: %r" % (block,))
      self.type = type_
      self.name = name
      self.uuid = uuid
      assert prevblock is None or isinstance(prevblock, _Block), \
          "not _Block: prevblock=%r" % (prevblock,)
      self._prev_dirent_blockref = prevblock
      if not isinstance(meta, Meta):
        M = Meta({'a': DEFAULT_DIR_ACL if self.isdir else DEFAULT_FILE_ACL})
        if meta is None:
          pass
        elif isinstance(meta, str):
          M.update_from_text(meta)
        else:
          raise ValueError("unsupported meta value: %r" % (meta,))
        if 'm' not in M:
          M['m'] = time.time()
        meta = M
      if type_ != DirentType.INDIRECT:
        self.meta = meta
      self.parent = parent

  def __repr__(self):
    return "%s:%s:%s(%r:%s,%r)" % (
        self.__class__.__name__,
        id(self),
        self.type,
        self.name,
        self.uuid,
        self.meta
    )

  @classmethod
  def from_str(cls, s, offset=0):
    ''' Parse a Dirent transcription from the str `s`.
    '''
    E, offset2 = parse_transcription(s, offset)
    if not isinstance(E, cls):
      raise ValueError(
          "expected instance of %s (got %s) at offset %d of %r"
          % (cls, type(E), offset, s))
    return E, offset2

  @staticmethod
  def from_components(type_, name, **kw):
    ''' Factory returning a _Dirent instance.
    '''
    if type_ == DirentType.DIR:
      cls = Dir
    elif type_ == DirentType.FILE:
      cls = FileDirent
    elif type_ == DirentType.SYMBOLIC:
      cls = SymlinkDirent
    elif type_ == DirentType.INDIRECT:
      cls = IndirectDirent
    else:
      warning("from_components: UNSUPPORTED TYPE %r, using _Dirent", type_)
      cls = partial(_Dirent, type_)
    return cls(name, **kw)

  # TODO: remove all uses of _Dirent.from_bytes
  @staticmethod
  def from_bytes(data, offset=0):
    ''' Factory to extract a Dirent from binary data at `offset` (default 0).
        Returns the Dirent and the new offset.
    '''
    return DirentRecord.value_from_bytes(data, offset=offset)

  @staticmethod
  def from_buffer(bfr):
    ''' Factory to extract a Dirent from the CornuCopyBuffer `bfr`.
        Returns the Dirent.
    '''
    return DirentRecord.value_from_buffer(bfr)

  def exists(self):
    ''' Does this exist? For a Dir this is always true: cogito ergo sum.
    '''
    return True

  def ingest_extended_data(self, extended_data):
    ''' The basic _Dirent subclasses do not use extended data.
    '''
    if extended_data:
      raise ValueError(
          "expected extended_data to be None or empty, got: %r"
          % (extended_data,))

  def get_extended_data(self):
    ''' The basic _Dirent subclasses do not use extended data.
    '''
    return None

  def __bytes__(self):
    ''' Serialise this Dirent to bytes.
    '''
    return bytes(DirentRecord(self))

  encode = __bytes__

  def __hash__(self):
    ''' Allows collecting _Dirents in a set.
    '''
    return id(self)

  def transcribe_inner(self, T, fp, attrs=None):
    ''' Transcribe the inner components of the Dirent as text.
    '''
    if attrs is None:
      attrs = {}
    if self.name and self.name != '.':
      T.transcribe(self.name, fp=fp)
      fp.write(':')
    if type(self) is _Dirent:
      attrs['type'] = self.type
    if self.uuid:
      attrs['uuid'] = self.uuid
    if self.type != DirentType.INDIRECT:
      if self.meta:
        attrs['meta'] = self.meta
      block = getattr(self, 'block', None)
      if block:
        attrs['block'] = block
    prev_blockref = self._prev_dirent_blockref
    if prev_blockref is not None:
      attrs['prevblock'] = prev_blockref
    T.transcribe_mapping(attrs, fp)

  @classmethod
  def parse_inner(cls, T, s, offset, stopchar, prefix):
    ''' Parse [name:]attrs from `s` at offset `offset`.
        Return _Dirent instance and new offset.
    '''
    name, offset2 = T.parse_qs(s, offset, optional=True)
    if name is None:
      name = ''
    else:
      if s[offset2] != ':':
        raise ValueError("offset %d: missing colon after name" % (offset2,))
      offset = offset2 + 1
    attrs, offset = T.parse_mapping(s, offset, stopchar)
    type_ = {
        'F': DirentType.FILE,
        'D': DirentType.DIR,
        'SymLink': DirentType.SYMBOLIC,
        'Indirect': DirentType.INDIRECT,
    }.get(prefix)
    return cls.from_components(type_, name, **attrs), offset

  def get_uuid(self):
    ''' Return this Dirent's UUID, creating it if necessary.
    '''
    u = self.uuid
    if u is None:
      u = self.uuid = uuid4()
      parent = self.parent
      if parent:
        parent.changed = True
    return u

  def pathto(self, R=None):
    ''' Return the path to this element if known.
        `R`: optional root Dirent, default None.
    '''
    E = self
    path = []
    while E is not R:
      path.append(E.name)
      E = E.parent
    return os.path.join(*reversed(path))

  def __eq__(self, other):
    block = getattr(self, 'block', None)
    oblock = getattr(other, 'block', None)
    return (
        self.name == other.name
        and self.type == other.type
        and self.meta == other.meta
        and (block is None if oblock is None else block == oblock)
    )

  @prop
  def prev_dirent(self):
    ''' Return the previous Dirent.

        If not None, during encoding or transcription, if self !=
        prev_dirent, include it in the encoding or transcription.

        TODO: parse out multiple blockrefs.
    '''
    prev_blockref = self._prev_dirent_blockref
    if prev_blockref is None:
      return None
    bfr = CornuCopyBuffer(prev_blockref.datafrom())
    E = _Dirent.from_buffer(bfr)
    if not bfr.at_eof():
      warning(
          "prev_dirent: _prev_dirent_blockref=%s:"
          " unparsed bytes after dirent at offset %d",
          prev_blockref, bfr.offset)
    return E

  @prev_dirent.setter
  @locked
  def prev_dirent(self, E):
    ''' Set the previous Dirent.
    '''
    assert isinstance(E, _Dirent), "set .prev_dirent: not a _Dirent: %s" % (E,)
    if E is None:
      if self._prev_dirent_blockref is not None:
        self.changed = True
    elif E == self:
      warning(
          "%r.prev_dirent=%s: ignore setting previous to our own state",
          self, E)
    else:
      Ebs = E.encode()
      self._prev_dirent_blockref = Block(data=Ebs)
      self.changed = True

  def snapshot(self):
    ''' Update the Dirent's previous block state if missing or changed.
    '''
    E = self.prev_dirent
    if E is None or E != self:
      self.prev_dirent = self

  def reconcile(self, other, force=False):
    ''' Reconcile 2 Dirents.

        It is expected that they are meant to be different revisions
        of the "same" object.
        If they both have UUIDs, they should match.
        They should be of the same type (both files, etc).
    '''
    with Pfx("%s.reconcile(%s)", self, other):
      if self is other:
        warning("tried to reconcile a _Dirent with itself")
        return
      uu1 = self.uuid
      uu2 = other.uuid
      if uu1 and uu2 and uu1 != uu2:
        warning("different UUIDs: %s vs %s", uu1, uu2)
        if not force:
          raise ValueError("not reconciling Dirents with different UUIDs")
      warning("NOT IMPLEMENTED YET")

  @property
  def isfile(self):
    ''' Is this a file _Dirent?
    '''
    return self.type == DirentType.FILE

  @property
  def isdir(self):
    ''' Is this a directory _Dirent?
    '''
    return self.type == DirentType.DIR

  @property
  def issym(self):
    ''' Is this a symbolic link _Dirent?
    '''
    return self.type == DirentType.SYMBOLIC

  @property
  def isindirect(self):
    ''' Is this an indirect _Dirent?
    '''
    return self.type == DirentType.INDIRECT

  def textencode(self):
    ''' Serialise the dirent as text.
    '''
    return totext(self.encode())

  @property
  def size(self):
    ''' Return this Dirent's length: its Block's span length.

        Note that Dirents with a None Block return None.
    '''
    block = getattr(self, 'block', None)
    return None if block is None else len(block)

  @property
  def mtime(self):
    ''' Return this Dirent's modification time (meta.mtime).
    '''
    return self.meta.mtime

  @mtime.setter
  def mtime(self, newtime):
    ''' Set this Dirent's modification time (meta.mtime).
    '''
    self.meta.mtime = newtime

  def touch(self, when=None):
    ''' Set the mtime for this file to `when` (default now: time.time()).
    '''
    if when is None:
      when = time.time()
    self.mtime = when

  def stat(self, fs=None):
    ''' Return this Dirent's POSIX stat structure.
    '''
    if fs is None:
      fs = defaults.fs
    M = self.meta
    I = fs.E2inode(self) if fs else None
    perm_bits = M.unix_perm_bits
    if perm_bits is None:
      if self.isdir:
        perm_bits = 0o700
      else:
        perm_bits = 0o600
    st_mode = self.unix_typemode | perm_bits
    st_ino = I.inum if I else -1
    # TODO: dev from FileSystem
    st_dev = fs.device_id
    if self.isdir:
      # TODO: should nlink for Dirs count its subdirs?
      st_nlink = 1
    else:
      st_nlink = I.refcount if I else 1
    st_uid = M.uid
    st_gid = M.gid
    if self.issym:
      pathref = self.pathref
      if pathref is None:
        warning("no pathref for %s", self)
        st_size = 0
      else:
        st_size = len(pathref)
    else:
      st_size = self.size
    st_atime = 0
    st_mtime = M.mtime
    st_ctime = 0
    return os.stat_result( (
        st_mode,
        st_ino,
        st_dev,
        st_nlink,
        st_uid, st_gid,
        st_size,
        st_atime, st_mtime, st_ctime,
    ) )

  @property
  def unix_typemode(self):
    ''' The portion of the mode bits defining the inode type.
    '''
    E = self
    if E.isindirect:
      E = E.ref
      if E.isindirect:
        raise ValueError(
            "indirect %s refers to another indirect: %s"
            % (self, E))
    if E.isdir:
      typemode = stat.S_IFDIR
    elif E.isfile:
      typemode = stat.S_IFREG
    elif E.issym:
      typemode = stat.S_IFLNK
    else:
      warning(
          "%s.unix_typemode: unrecognised type %d, pretending S_IFREG"
          % (type(self), self.type))
      typemode = stat.S_IFREG
    return typemode

register_transcriber(_Dirent, (
    'INVALIDDirent',
    'SymLink',
    'Indirect',
    'D',
    'F',
))

class InvalidDirent(_Dirent):
  ''' Encapsulation for an invalid Dirent data chunk.
  '''

  transcribe_prefix = 'INVALIDDirent'

  def __init__(self, name, *, chunk=None, **kw):
    ''' An invalid Dirent.
        Record the original data chunk for regurgitation later.
    '''
    _Dirent.__init__(
        self,
        DirentType.INVALID,
        name,
        **kw)
    self.chunk = chunk

  def __str__(self):
    return '<InvalidDirent:%s>' % (texthexify(self.chunk),)

  def encode(self):
    ''' Return the original data chunk.
    '''
    return self.chunk

  def transcribe_inner(self, T, fp):
    ''' Transcribe the inner components of this InvalidDirent's transcription.
    '''
    return super().transcribe_inner(T, fp, {'chunks': self.chunk})

class SymlinkDirent(_Dirent):
  ''' A symbolic link.
  '''

  transcribe_prefix = 'SymLink'

  def __init__(self, name, *, target=None, **kw):
    super().__init__(DirentType.SYMBOLIC, name, **kw)
    if target is None:
      if self.meta.pathref is None:
        raise ValueError("missing target")
    else:
      self.meta.pathref = target

  @property
  def pathref(self):
    ''' The symbolic link's path reference.
    '''
    return self.meta.pathref

  def transcribe_inner(self, T, fp):
    ''' Transcribe the inner components for a SymlinkDirent.
    '''
    return super().transcribe_inner(T, fp, {})

class IndirectDirent(_Dirent):
  ''' An indirect Dirent, referring to another Dirent by UUID.

      This is how a feature like a hard link is implented in a vt filesystem.

      *Note*: unlike other Dirents, IndirectDirents are considered
      emphemeral, specificly in that their uuid attribute is a
      reference to another persistent Direct. Obtaining the target
      dirent requires dereferencing through a FileSystem.
  '''

  transcribe_prefix = 'Indirect'

  def __init__(self, name, uuid, meta=None, block=None):
    if block is not None:
      raise ValueError("IndirectDirent block should be None, got: %r" % (block,))
    if meta is not None:
      raise ValueError("IndirectDirent meta should be None, got: %r" % (meta,))
    _Dirent.__init__(self, DirentType.INDIRECT, name, uuid=uuid)

  def deref(self, fs=None):
    ''' Dereference this IndirectDirent's UUID via a FileSystem.
    '''
    if fs is None:
      fs = defaults.fs
      if not fs:
        error("NO CURRENT FILESYSTEM")
        stack_dump()
        raise ValueError("no current FileSystem")
    try:
      I = fs[self.uuid]
    except KeyError as e:
      error("%s: no inode for UUID %s: %s", fs, self.uuid, e)
      stack_dump()
      raise
    return I.E

  @prop
  def ref(self):
    ''' The referenced Dirent via the default FileSystem.
    '''
    try:
      return self.deref()
    except KeyError as e:
      raise AttributeError('ref') from e

  @property
  def meta(self):
    ''' The metadata of the referenced Dirent.
    '''
    return self.ref.meta

  @meta.setter
  def meta(self, new_meta):
    ''' Setting the metadata acts on the referent.
    '''
    self.ref.meta = new_meta

  @prop
  def block(self):
    ''' The content block for the referenced Dirent.
    '''
    return self.ref.block

  def transcribe_inner(self, T, fp):
    ''' Transcribe the inner components of an IndirectDirent.
    '''
    return super().transcribe_inner(T, fp, {})

class FileDirent(_Dirent, MultiOpenMixin, FileLike):
  ''' A _Dirent subclass referring to a file.

      If closed, ._block refers to the file content.
      If open, .open_file refers to the content.

      NOTE: multiple opens initialise the _same_ backing file, with a shared
      read/write offset. File systems must share this, and maintain their own
      offsets in their file handle objects.
  '''

  transcribe_prefix = 'F'

  def __init__(self, name, block=None, **kw):
    _Dirent.__init__(self, DirentType.FILE, name, **kw)
    MultiOpenMixin.__init__(self)
    if block is None:
      block = Block(data=b'')
    self.open_file = None
    self._block = block
    self._check()

  # FileLike.lstat uses the common stat method
  lstat = stat

  @locked
  def startup(self):
    ''' Set up .open_file on first open.
    '''
    self._check()
    if self.open_file is not None:
      raise RuntimeError("first open, but .open_file is not None: %r" % (self.open_file,))
    if self._block is None:
      raise RuntimeError("first open, but ._block is None")
    self.open_file = RWBlockFile(self._block)
    self._block = None
    self._check()

  @locked
  def shutdown(self):
    ''' On final close, close .open_file and save result as ._block.
    '''
    self._check()
    if self._block is not None:
      error(
          "final close, but ._block is not None;"
          " replacing with self.open_file.close(), was: %s",
          self._block)
    f = self.open_file
    f.filename = self.name
    self._block = f.sync()
    f.close()
    self.open_file = None
    self._check()

  def _check(self):
    ''' Internal consistency check.
    '''
    # TODO: check ._block and .open_file against MultiOpenMixin open count
    if self._block is None:
      if self.open_file is None:
        raise ValueError("both ._block and .open_file are None")
    ## both are allowed to be set
    ##elif self.open_file is not None:
    ##  raise ValueError("._block is %s and .open_file is %r" % (self._block, self.open_file))

  @classmethod
  def from_chunks(cls, chunks):
    ''' Create a FileDirent from an iterable of bytes `chunks`.
    '''
    return cls('', block=top_block_for(blockify(chunks)))

  @property
  def block(self):
    ''' Obtain the top level Block.
        If open, sync the file to update ._block.
    '''
    self._check()
    ##X("access FileDirent.block from:")
    ##stack_dump(indent=2)
    with self._lock:
      if self.open_file is None:
        return self._block
    return self.open_file.sync()

  @block.setter
  @locked
  def block(self, B):
    ''' Update the Block for this FileDirent.
        The Dirent is expected to be closed.
    '''
    self._check()
    if self.open_file is not None:
      raise RuntimeError("tried to set .block directly while open")
    self._block = B

  def datafrom(self):
    ''' Generator yielding data from the FileDirent.
    '''
    f = self.open_file
    if f is None:
      return self.block.datafrom()
    return f.datafrom()

  @property
  def size(self):
    ''' Return the size of this file.
        If open, use the open file's size.
        Otherwise get the length of the top Block.
    '''
    if self.open_file is None:
      sz = len(self.block)
    else:
      sz = len(self.open_file)
    return sz

  def flush(self, scanner=None, dispatch=None):
    ''' Flush the contents of the file.
        Presumes the Dirent is open.
    '''
    return self.open_file.flush(scanner, dispatch=dispatch)

  def truncate(self, length):
    ''' Truncate this FileDirent to the specified size.
    '''
    Esize = self.size
    if Esize != length:
      with self:
        return self.open_file.truncate(length)
    return None

  # TODO: move into distinctfile utilities class with rsync-like stuff etc
  def restore(self, path, makedirs=False, verbosefp=None):
    ''' Restore this _Dirent's file content to the name `path`.
    '''
    with Pfx("FileDirent.restore(%s)", path):
      if verbosefp is not None:
        verbosefp.write(path)
        verbosefp.write('\n')
      dirpath = os.path.dirname(path)
      if dirpath and not os.path.isdir(dirpath):
        if makedirs:
          os.makedirs(dirpath)
      with open(path, "wb") as ofp:
        for B in self.block.leaves():
          ofp.write(B.blockdata())
        fd = ofp.fileno()
        st = os.fstat(fd)
        user, group, perms = self.meta.unixPerms()
        if user is not None or group is not None:
          os.fchmod(fd, perms)
        if user is None:
          uid = -1
        else:
          uid = pwd.getpwnam(user)[2]
          if uid == st.st_uid:
            uid = -1
        if group is None:
          gid = -1
        else:
          gid = grp.getgrnam(group)[2]
          if gid == st.st_gid:
            gid = -1
        if uid != -1 or gid != -1:
          os.fchown(fd, uid, gid)
      if self.meta.mtime is not None:
        os.utime(path, (st.st_atime, self.meta.mtime))

  def transcribe_inner(self, T, fp):
    ''' Transcribe the inner components of this FileDirent's transcription.
    '''
    return _Dirent.transcribe_inner(self, T, fp, {})

  def pushto(self, S2, Q=None, runstate=None):
    ''' Push the Block with the file contents to the Store `S2`.

        Parameters:
        * `S2`: the secondary Store to receive Blocks
        * `Q`: optional preexisting Queue, which itself should have
          come from a .pushto targetting the Store `S2`.
        * `runstate`: optional RunState used to cancel operation

        Semantics are as for cs.vt.block.Block.pushto.
    '''
    return self.block.pushto(S2, Q=Q, runstate=runstate)

class Dir(_Dirent, DirLike):
  ''' A directory.

      Special attributes:
      * `changed`:
        Starts False, becomes true if this or any subdirectory gets changed
        or has a file opened; stays True from then on.
        This accepts an ongoing compute cost for .block to avoid
        setting the flag on every file.write etc.
  '''

  transcribe_prefix = 'D'

  def __init__(self, name, block=None, **kw):
    ''' Initialise this directory.

        Parameters:
        * `meta`: meta information
        * `parent`: parent Dir
        * `block`: pre-existing Block with initial Dir content
    '''
    _Dirent.__init__(self, DirentType.DIR, name, **kw)
    if block is None:
      self._block = None
      self._entries = {}
    else:
      self._block = block
      self._entries = None
    self._unhandled_dirent_chunks = None
    self._changed = False
    self._change_notifiers = None
    self._lock = RLock()

  # DirLike.lstat uses the common stat method
  lstat = stat

  @prop
  def changed(self):
    ''' Whether this Dir has been changed.
    '''
    return self._changed

  @changed.setter
  @locked
  def changed(self, status):
    ''' Mark this dirent as changed or not changed;
        propagate truth to parent Dir if present.
    '''
    if not status:
      raise ValueError("cannot clear .changed")
    E = self
    while E is not None:
      E._changed = True
      E._notify_change()
      E = E.parent

  @locked
  def on_change(self, notifier):
    ''' Record the callable `notifier` to fire when .changed is set.
    '''
    notifiers = self._change_notifiers
    if notifiers is None:
      self._change_notifiers = notifiers = set()
    notifiers.add(notifier)

  def _notify_change(self):
    ''' Call each recorded notifier with this Dir.
    '''
    notifiers = self._change_notifiers
    if notifiers:
      for notifier in notifiers:
        notifier(self)

  @property
  @locked
  def entries(self):
    ''' Property containing the live dictionary holding the Dir entries.
    '''
    emap = self._entries
    if emap is None:
      # compute the dictionary holding the live Dir entries
      emap = {}
      for E in DirentRecord.parse_buffer_values(self._block.bufferfrom()):
        E.parent = self
        emap[E.name] = E
      self._entries = emap
    return emap

  @property
  def size(self):
    ''' The length of a Dir is the number of entries it contains.
        This property is used mostly for stat calls.
    '''
    return len(self.entries)

  @property
  @locked
  @logexc
  def block(self):
    ''' Return the top Block referring to an encoding of this Dir.

        TODO: blockify the encoding? Probably desirable for big Dirs.
    '''
    if self._block is None or self.changed:
      warning(
          "Dir(%d:%r): recompute block: current _block=%s, changed=%s ...",
          id(self), self.name, self._block, self.changed)
      stack_dump()
      # recompute in case of change
      # restore the unparsed Dirents from initial load
      if self._unhandled_dirent_chunks is None:
        data = b''
      else:
        data = b''.join(self._unhandled_dirent_chunks)
      # append the valid or new Dirents
      names = sorted(self.keys())
      data += b''.join(
          self[name].encode()
          for name in names
          if name != '.' and name != '..'
      )
      # TODO: if len(data) >= 16384 blockify?
      B = self._block = Block(data=data)
      self._changed = False
    else:
      B = self._block
    return B

  def dirs(self):
    ''' Return a list of the names of subdirectories in this Dir.
    '''
    return [ name for name in self.keys() if self[name].isdir ]

  def files(self):
    ''' Return a list of the names of files in this Dir.
    '''
    return [ name for name in self.keys() if self[name].isfile ]

  @staticmethod
  def _validname(name):
    ''' Test if a name is valid: not empty and not containing the path separator.
    '''
    return len(name) > 0 and name.find(PATHSEP) < 0

  def get(self, name, dflt=None):
    ''' Fetch the Dirent named `name` or `dflt`.
    '''
    try:
      E = self[name]
    except KeyError:
      return dflt
    return E

  def keys(self):
    ''' Return the Dirent names contained in this Dir. (Mapping method.)
    '''
    return self.entries.keys()

  def items(self):
    ''' Return the Dirents contained in this Dir. (Mapping method.)
    '''
    return self.entries.items()

  def __contains__(self, name):
    if name == '.':
      return True
    if name == '..':
      return self.parent is not None
    return name in self.entries

  def __iter__(self):
    return iter(self.keys())

  def __getitem__(self, name):
    if name == '.':
      return self
    if name == '..':
      return self.parent
    return self.entries[name]

  def __setitem__(self, name, E):
    ''' Store a _Dirent in the specified name slot.
    '''
    if not self._validname(name):
      raise KeyError("invalid name: %s" % (name,))
    if not isinstance(E, _Dirent):
      raise ValueError("E is not a _Dirent: <%s>%r" % (type(E), E))
    self.entries[name] = E
    self.touch()
    self.changed = True
    E.name = name
    Eparent = E.parent
    if Eparent is None:
      E.parent = self
    elif Eparent is not self:
      warning("%s: changing %r.parent to self, was %s", self, name, Eparent)
      E.parent = self

  def pop(self, name):
    ''' Delete `name` and return the Dirent.
    '''
    if not self._validname(name):
      raise KeyError("invalid name: %s" % (name,))
    if name == '.' or name == '..':
      raise KeyError("refusing to delete . or ..: name=%s" % (name,))
    E = self.entries[name]
    del self.entries[name]
    E.parent = None
    self.touch()
    self.changed = True
    return E

  __delitem__ = pop

  def add(self, E):
    ''' Add a Dirent to this Dir.
        If the name is already taken, raise KeyError.
    '''
    name = E.name
    if name in self:
      raise KeyError("name already exists: %r" % (name,))
    self[name] = E

  @locked
  def rename(self, oldname, newname):
    ''' Rename entry `oldname` to entry `newname`.
    '''
    E = self[oldname]
    del E[oldname]
    E.name = newname
    self[newname] = E

  def mkdir(self, name):
    ''' Create a subdirectory named `name`, return the _Dirent.
    '''
    debug("<%s>.mkdir(%s)..." % (self.name, name))
    D = self[name] = Dir(name, parent=self)
    return D

  def file_frombuffer(self, name, bfr):
    if name in self:
      raise KeyError("name already exists: %r" % (name,))
    E = FileDirent.from_chunks(iter(bfr))
    self[name] = E
    return E

  def chdir1(self, name):
    ''' Change directory to the immediate entry `name`.
        Return the entry.
    '''
    D = self[name]
    if not D.isdir:
      raise ValueError("%s[name=%s]: not a directory" % (self, name))
    return D

  def chdir(self, path):
    ''' Change directory to `path`, return the ending directory.
    '''
    D = self
    for name in path.split(PATHSEP):
      if name:
        D = D.chdir1(name)
    return D

  def makedirs(self, path, force=False):
    ''' Like os.makedirs(), create a directory path at need; return the bottom Dir.
        If `force`, replace any non-Dir encountered with an empty Dir.
    '''
    E = self
    if isinstance(path, str):
      subpaths = path_split(path)
    else:
      subpaths = path
    for name in subpaths:
      if name == '' or name == '.':
        continue
      if name == '..':
        E = E.parent
        continue
      subE = E.get(name)
      if subE is None:
        subE = E.mkdir(name)
      else:
        if not subE.isdir:
          if force:
            subE = E.mkdir(name)
          else:
            raise ValueError("%s[name=%s] is not a directory" % (subE, name))
      E = subE
    return E

  def new_name(self, prefix, n=1):
    ''' Allocate a new unused name with the supplied `prefix`.
    '''
    while True:
      name2 = '.'.join(prefix, str(n))
      if name2 not in self:
        return name2
      n += 1

  def update(self, D2, path=None):
    ''' Update this Dir with changes from `D2` which is presumed to be new.
        Note: this literally attaches nodes from `D2` into this
        Dir's tree where possible.

        TODO: replace with code from vt.merge.
    '''
    if path is None:
      path = self.pathto()
    with Pfx("update(%r)", path):
      for name in D2:
        with Pfx(name):
          E2 = D2[name]
          if name in self:
            # conflict
            # TODO: support S_IFWHT whiteout entries
            E1 = self[name]
            if E1.uuid == E2.uuid:
              # same file
              assert E1.type == E2.type
              if E2.meta.ctime > E1.meta.ctime:
                info("update meta")
                E1.meta.update(E2.meta.items())
              if E1.block != E2.block:
                if E2.mtime > E1.mtime:
                  # TODO: E1.flush _after_ backend update? or before?
                  info("update block => %s", E2.block)
                  E1.block = E2.block
                  E1.meta.mtime = E2.mtime
            else:
              # distinct objects
              if E1.isdir and E2.isdir:
                # merge subtrees
                E1.update(E2)
              elif E1.isfile and E2.isfile and E1.block == E2.block:
                # file with same content, fold
                # TODO: use Block.compare_content if different blocks
                if E2.meta.ctime > E1.meta.ctime:
                  info("update meta")
                  E1.meta.update(E2.meta.items())
              else:
                # different content
                # add other object under a different name
                new_name = self.new_name(name)
                info("add new entry as %r: %s", new_name, E2)
                self[new_name] = E2
          else:
            # new item
            # NB: we don't recurse into new Dirs, not needed
            info("add new entry: %s", E2)
            self[name] = E2

  def update_notification(self, newE, when, source):
    ''' Receptor for Archive.update() notifications.
    '''
    with Pfx("update_notification(%s,when=%s,source=%s)", newE, when, source):
      if newE is not self:
        self.update(newE)

  def transcribe_inner(self, T, fp):
    return _Dirent.transcribe_inner(self, T, fp, {})

  def pushto(self, S2, Q=None, runstate=None):
    ''' Push the Dir Blocks to the Store `S2`.

        Parameters:
        * `S2`: the secondary Store to receive Blocks
        * `Q`: optional preexisting Queue, which itself should have
          come from a .pushto targetting the Store `S2`.
        * `runstate`: optional RunState used to cancel operation

        This pushes the Dir's Block encoding to `S2` and then
        recursively pushes each Dirent's Block data to `S2`.
    '''
    if Q is None:
      # create a Queue and a worker Thread
      Q, T = defaults.S.pushto(S2)
    else:
      # use an existing Queue, no Thread to wait for
      T = None
    B = self.block
    # push the Dir block data
    B.pushto(S2, Q=Q, runstate=runstate)
    # and recurse into contents
    for E in DirentRecord.parse_buffer_values(B.bufferfrom()):
      if runstate and runstate.cancelled:
        warning("pushto(%s) cancelled", self)
        break
      E.pushto(S2, Q=Q, runstate=runstate)
    if T:
      Q.close()
      T.join()

if __name__ == '__main__':
  from .dir_tests import selftest
  selftest(sys.argv)



#!/usr/bin/python
#
# Processes for pushing or pulling blocks between Stores.
#   - Cameron Simpson <cs@cskk.id.au> 13dec2015
# 

from functools import partial
import sys
from cs.result import OnDemandFunction
from cs.x import X

def pull_hashcode(S1, S2, hashcode):
  ''' Fetch the data for `hashcode` from `S2` and store in `S1`; return the data.
  '''
  data = S2.fetch(hashcode)
  h1 = S1.add(data)
  if h1 != hashcode:
    raise ValueError("%s.add(%s.fetch(%s)) gave different hashcode: %s"
                     % (S1, S2, hashcode, h1))
  return data

def pull_hashcodes(S1, S2, hashcodes):
  ''' Generator which fetches the data for the supplied `hashcodes` from `S2` if not in `S1`, updating `S1`.
      The generator yields (hashcode, get_data) where hashcode is
      each hashcode and get_data is a callable with returns the
      data.
  '''
  # mapping from hashcode to Result
  fetching = {}
  LFs = []
  for hashcode in hashcodes:
    if hashcode in fetching:
      yield hashcode, fetching[hashcode]
    elif hashcode in S1:
      yield hashcode, OnDemandFunction(S1.fetch, hashcode)
    else:
      # initiate pull of hashcode via S2's scheduler
      # TODO: purge hashcode from fetching after pull in race free fashion
      # to that we can run this indefinitely
      LF = fetching[hashcode] = S2._defer(pull_hashcode, S1, S2, hashcode)
      yield hashcode, LF

def missing_hashcodes(S1, S2, window_size=None):
  ''' Scan Stores `S1` and `S2` and yield hashcodes in `S2` but not in `S1`.
      This relies on both Stores supporting the .hashcodes method;
      dumb unordered Stores do not.
      `window_size`: number of hashcodes to fetch at a time for comparison,
                     default 1024.
  '''
  if window_size is None:
    window_size = 1024
  hashcodes1 = None
  hashcodes2 = list(S2.hashcodes(length=window_size))
  while hashcodes2 and (hashcodes1 is None or hashcodes1):
    # note end of S2 window so that we can fetch the next window
    last_hashcode2 = hashcodes2[-1]
    # test each S2 hashcode for presence in S1
    for ndx, hashcode in enumerate(hashcodes2):
      if hashcodes1 is None or hashcode > last_hashcode1:
        # past the end of S1 hashcode window
        # fetch new window starting at current hashcode
        hashcodes1 = list(S1.hashcodes(start_hashcode=hashcode, length=window_size))
        if not hashcodes1:
          # no more S1 hashcodes, cease scan
          # keep the unscanned hashcodes (including current)
          # to yield in the post-scan loop, where we don't bother consulting S1
          hashcodes2 = hashcodes2[ndx:]
          break
        # note highest hashcode as top of window
        last_hashcode1 = hashcodes1[-1]
        # convert to set for fast membership lookup
        hashcodes1 = set(hashcodes1)
      if hashcode not in hashcodes1:
        yield hashcode
    if not hashcodes1:
      break
    # fetch next batch of hashcodes from S2
    hashcodes2 = list(S2.hashcodes(start_hashcode=last_hashcode2, length=window_size,
                                   after=True))
  # no more hashcodes in S1 - gather everything else in S2
  while True:
    hashcode = None
    for hashcode in hashcodes2:
      yield hashcode
    if hashcode is None:
      break
    # fetch next bunch of hashcodes
    hashcodes2 = S2.hashcodes(start_hashcode=hashcode, length=window_size, after=True)

def missing_hashcodes_by_checksum(S1, S2, window_size=None):
  ''' Scan Stores `S1` and `S2` and yield hashcodes in `S2` but not in `S1`.
      This relies on both Stores supporting the .hashcodes and
      .hash_of_hashcodes methods; dumb unordered Stores do not.
      `window_size`: intial number of hashcodes to fetch at a time for comparison,
                     default 1024.
  '''
  if window_size is None:
    window_size = 1024
  # latest hashcode already compared
  start_hashcode = None
  after = False
  while True:
    # collect checksum of hashcodes after start_hashcode from S1 and S2
    hash1, h_final1 = S1.hash_of_hashcodes(length=window_size, start_hashcode=start_hashcode, after=after)
    if h_final1 is None:
      # end of S1 hashcodes - return all following S2 hashcodes
      break
    hash2, h_final2 = S2.hash_of_hashcodes(length=window_size, start_hashcode=start_hashcode, after=after)
    if h_final2 is None:
      # end of S2 hashcodes - done - return from function
      return
    if hash1 == hash2:
      if h_final1 != h_final2:
        raise RuntimeError('hashes match but h_final1=%s != h_final2=%s'
                           % (h_final1, h_final2))
      # this chunk matches, fetch the next
      start_hashcode = h_final1
      after = True
      continue
    # mismatch, try smaller window
    if window_size >= 32:
      # shrink window until match found or window too small to bother
      owindow_size = window_size
      window_size //= 2
      continue
    # fetch the actual hashcodes
    hashcodes2 = list(S2.hashcodes(start_hashcode=start_hashcode, length=window_size, after=after))
    if not hashcodes2:
      # maybe some entires removed? - anyway, no more S2 so return
      return
    hashcodes1 = set(S1.hashcodes(start_hashcode=start_hashcode, length=window_size, after=after))
    if not hashcodes1:
      # maybe some entries removed? - anyway, no more S1 so return all following S2 hashcodes
      break
    # in case things changed since earlier checksum
    h_final1 = max(hashcodes1)
    for ndx, hashcode in enumerate(hashcodes2):
      # note that if we fetch more hashcodes then we may get am empty window
      # just keep running with that if so - do not check h_final1
      if hashcodes1 and h_final1 < hashcode:
        # hashcodes1 does not cover this point in hashcodes2, fetch more
        hashcodes1 = set(S1.hashcodes(start_hashcode=hashcode, length=len(hashcodes2)-ndx))
        if hashcodes1:
          h_final1 = max(hashcodes1)
        else:
          h_final = None
      if hashcode not in hashcodes1:
        yield hashcode
    # resume scan from here
    start_hashcode = hashcodes2[-1]
    after = True
  # collect all following S2 hashcodes
  while True:
    hashcodes2 = list(S2.hashcodes(start_hashcode=start_hashcode, length=window_size, after=after))
    if not hashcodes2:
      break
    for hashcode in hashcodes2:
      yield hashcode
    start_hashcode = hashcodes2[-1]
    after = True

if __name__ == '__main__':
  from .pushpull_tests import selftest
  selftest(sys.argv)



#!/usr/bin/env python3
#
# Command script for venti-ish implementation.
#       - Cameron Simpson <cs@cskk.id.au> 01may2007
#

''' cs.vt command line utility.
'''

from __future__ import with_statement
from collections import defaultdict
from datetime import datetime
import errno
from getopt import getopt, GetoptError
import logging
import os
from os.path import basename, realpath, splitext, expanduser, \
    exists as pathexists, join as joinpath, \
    isdir as isdirpath, isfile as isfilepath
import shutil
from signal import signal, SIGINT, SIGHUP, SIGQUIT
import sys
from threading import Thread
from time import sleep
from cs.debug import ifdebug, dump_debug_threads, thread_dump
from cs.env import envsub
from cs.fileutils import file_data, shortpath
from cs.lex import hexify, get_identifier
import cs.logutils
from cs.logutils import exception, error, warning, info, debug, \
                        setup_logging, logTo, loginfo
from cs.pfx import Pfx
from cs.resources import RunState
from cs.tty import statusline, ttysize
import cs.x
from cs.x import X
from . import fromtext, defaults, DEFAULT_CONFIG_PATH
from .archive import Archive, CopyModes
from .block import BlockRecord
from .blockify import blocked_chunks_of
from .compose import get_store_spec
from .config import Config, Store
from .convert import expand_path
from .datadir import DataDirIndexEntry
from .datafile import DataFileReader
from .debug import dump_chunk, dump_Block
from .dir import Dir
from .fsck import fsck_Block, fsck_dir
from .hash import DEFAULT_HASHCLASS
from .index import LMDBIndex
from .merge import merge
from .parsers import scanner_from_filename
from .paths import OSDir, OSFile, decode_Dirent_text, dirent_dir, dirent_file, dirent_resolve
from .server import serve_tcp, serve_socket
from .store import ProgressStore, ProxyStore
from .transcribe import parse

def main(*a, **kw):
  ''' Create a VTCmd instance and call its main method.
  '''
  return VTCmd().main(*a, **kw)

class VTCmd:
  ''' A main programme instance.
  '''

  USAGE = '''Usage: %s [option...] [profile] subcommand [arg...]
  Options:
    -C store  Specify the store to use as a cache.
              Specify "NONE" for no cache.
              Default: from $VT_CACHE_STORE or "[cache]".
    -S store  Specify the store to use:
                [clause]        Specification from .vtrc.
                /path/to/dir    GDBMStore
                tcp:[host]:port TCPStore
                |sh-command     StreamStore via sh-command
              Default from $VT_STORE, or "[default]", except for
              the "serve" subcommand which defaults to "[server]"
              and ignores $VT_STORE.
    -f config Config file. Default from $VT_CONFIG, otherwise ''' \
    + DEFAULT_CONFIG_PATH + '''
    -q        Quiet; not verbose. Default if stderr is not a tty.
    -v        Verbose; not quiet. Default if stderr is a tty.
  Subcommands:
    cat filerefs...
    config
    dump {datafile.vtd|index.gdbm|index.lmdb}
    fsck block blockref...
    import [-oW] path {-|archive.vt}
    init
    ls [-R] dirrefs...
    mount [-a] [-o {append_only,readonly}] [-r] {Dir|config-clause|archive.vt} [mountpoint [subpath]]
      -a  All dates. Implies readonly.
      -o options
          Mount options:
            append_only Files may not be truncated or overwritten.
            readonly    Read only; data may not be modified.
      -r  Readonly, the same as "-o readonly".
    pack path
    pullfrom other-store objects...
    pushto other-store objects...
    serve [{DEFAULT|-|/path/to/socket|host:port} [name:storespec]...]
    test blockify file
    unpack archive.vt
'''

  def main(self, argv=None, environ=None, verbose=None, subcmd=None):
    ''' The main function for cs.vt.

        Parameters:
        * `argv`: the command line arguments,
          default from `sys.argv`.
        * `environ`: the environment variable mapping,
          default from `os.environ`.
        * `verbose`: verbose mode, also activated by the `-v` option.
        * `subcmd`: which subcommand to run,
          default from the first argument after the options.
    '''
    global loginfo
    if argv is None:
      argv = sys.argv
    self.argv = argv
    if environ is None:
      environ = os.environ

    if verbose is None:
      # verbose if stderr is a tty
      try:
        verbose = sys.stderr.isatty()
      except AttributeError:
        verbose = False
    self.verbose = verbose

    cmd = basename(argv[0])
    args = argv[1:]
    if cmd.endswith('.py'):
      cmd = 'vt'
    self.cmd = cmd

    usage = self.USAGE % (cmd,)

    badopts = False

    setup_logging(cmd_name=cmd, upd_mode=sys.stderr.isatty(), verbose=self.verbose)
    ####cs.x.X_logger = logging.getLogger()

    config_path = os.environ.get('VT_CONFIG', expanduser(DEFAULT_CONFIG_PATH))
    store_spec = None
    cache_store_spec = os.environ.get('VT_CACHE_STORE', '[cache]')
    dflt_log = os.environ.get('VT_LOGFILE')

    try:
      opts, args = getopt(args, 'C:S:f:qv')
    except GetoptError as e:
      error("unrecognised option: %s: %s", e.opt, e.msg)
      badopts = True
      opts, args = [], []

    for opt, val in opts:
      if opt == '-C':
        if val == 'NONE':
          cache_store_spec = None
        else:
          cache_store_spec = val
      elif opt == '-S':
        # specify Store
        store_spec = val
      elif opt == '-f':
        config_path = val
      elif opt == '-q':
        # quiet: not verbose
        self.verbose = False
      elif opt == '-v':
        # verbose: not quiet
        self.verbose = True
      else:
        raise RuntimeError("unhandled option: %s" % (opt,))

    self.config_path = config_path
    self.store_spec = store_spec
    self.cache_store_spec = cache_store_spec

    if self.verbose:
      loginfo.level = logging.INFO
      upd = loginfo.upd
      if upd is not None:
        upd.nl_level = logging.INFO

    if dflt_log is not None:
      logTo(dflt_log, delay=True)

    xit = None
    self.runstate = RunState("main")
    with defaults.push_runstate(self.runstate):
      self.config = Config(self.config_path)

      # catch signals, flag termination
      def sig_handler(sig, frame):
        ''' Signal handler
        '''
        warning("received signal %s from %s", sig, frame)
        if sig == SIGQUIT:
          thread_dump()
        X("%s.cancel()...", self.runstate)
        self.runstate.cancel()
        if sig == SIGQUIT:
          sys.exit(1)
      signal(SIGHUP, sig_handler)
      signal(SIGINT, sig_handler)
      signal(SIGQUIT, sig_handler)

      try:
        xit = self.cmd_op(args, op=subcmd)
      except GetoptError as e:
        error("%s", e)
        badopts = True

      if badopts:
        sys.stderr.write(usage)
        return 2

      if not isinstance(xit, int):
        raise RuntimeError("exit code not set by subcommand: %r" % (xit,))

      if ifdebug():
        dump_debug_threads()

    return xit

  def cmd_op(self, args, op=None):
    ''' Run a subcommand from `args`.
    '''
    if op is None:
      try:
        op = args[0]
      except IndexError:
        raise GetoptError("missing command")
      args = args[1:]
    with Pfx(op):
      if op == "profile":
        return self.cmd_profile(args)
      try:
        op_func = getattr(self, "cmd_" + op)
      except AttributeError:
        raise GetoptError("unknown subcommand \"%s\"" % (op,))
      # these commands run without a context Store
      if op in ("config", "dump", "init", "scan", "test"):
        return op_func(args)
      # open the default Store
      if self.store_spec is None:
        if op == "serve":
          store_spec = '[server]'
        else:
          store_spec = os.environ.get('VT_STORE', '[default]')
        self.store_spec = store_spec
      try:
        # set up the primary Store using the main programme RunState for control
        S = Store(self.store_spec, self.config, runstate=self.runstate)
      except ValueError as e:
        raise GetoptError("unusable Store specification: %s: %s" % (self.store_spec, e))
      except Exception as e:
        exception("UNEXPECTED EXCEPTION: can't open store %r: %s", self.store_spec, e)
        raise GetoptError("unusable Store specification: %s" % (self.store_spec,))
      defaults.push_Ss(S)
      if self.cache_store_spec is None:
        cacheS = None
      else:
        try:
          cacheS = Store(self.cache_store_spec, self.config)
        except Exception as e:
          exception("can't open cache store %r: %s", self.cache_store_spec, e)
          raise GetoptError(
              "unusable Store specification: %s"
              % (self.cache_store_spec,))
        else:
          S = ProxyStore(
              "%s:%s" % (cacheS.name, S.name),
              read=(cacheS,),
              read2=(S,),
              copy2=(cacheS,),
              save=(cacheS, S)
          )
          S.config = self.config
      ##X("MAIN CMD_OP S:")
      ##dump_Store(S)
      defaults.push_Ss(S)
      # start the status ticker
      if False and sys.stdout.isatty():
        X("wrap in a ProgressStore")
        run_ticker = True
        S = ProgressStore("ProgressStore(%s)" % (S,), S)
        def ticker():
          old_text = ''
          while run_ticker:
            text = S.status_text()
            if text != old_text:
              statusline(text)
              old_text = text
            sleep(0.25)
        T = Thread(name='%s-status-line' % (S,), target=ticker)
        T.daemon = True
        T.start()
      else:
        run_ticker = False
      with S:
        xit = op_func(args)
      if cacheS:
        cacheS.backend = None
      if run_ticker:
        run_ticker = False
      return xit

  def cmd_profile(self, *a, **kw):
    ''' Wrapper to profile other subcommands and report.
    '''
    try:
      import cProfile as profile
    except ImportError:
      import profile
    P = profile.Profile()
    P.enable()
    try:
      xit = self.cmd_op(*a, **kw)
    except Exception:
      P.disable()
      raise
    P.disable()
    P.create_stats()
    P.print_stats(sort='cumulative')
    return xit

  def cmd_cat(self, args):
    ''' Concatentate the contents of the supplied filerefs to stdout.
    '''
    if not args:
      raise GetoptError("missing filerefs")
    for path in args:
      cat(path)
    return 0

  def cmd_config(self, args):
    ''' Recite the configuration.
    '''
    if args:
      raise GetoptError("extra arguments: %r" % (args,))
    self.config.write(sys.stdout)
    return 0

  def cmd_dump(self, args):
    ''' Dump various file types.
    '''
    if not args:
      raise GetoptError("missing filerefs")
    hashclass = DEFAULT_HASHCLASS
    one_line = True
    _, columns = ttysize(1)
    if columns is None:
      columns = 80
    max_width = columns - 1
    for path in args:
      if path.endswith('.vtd'):
        print(path)
        DF = DataFileReader(path)
        with DF:
          try:
            for DR in DF.scanfrom(0):
              data = DR.data
              hashcode = hashclass(data)
              leadin = '%9d %16.16s' % (offset, hashcode)
              dump_chunk(data, leadin, max_width, one_line)
          except EOFError:
            pass
      elif path.endswith('.lmdb'):
        print(path)
        lmdb = LMDBIndex(path[:-5], hashclass, decode=DataDirIndexEntry.from_bytes)
        with lmdb:
          for hashcode, entry in lmdb.items():
            print(hashcode, entry)
      else:
        warning("unsupported file type: %r", path)
    return 0

  def cmd_fsck(self, args):
    ''' Data structure inspection/repair.
    '''
    if not args:
      raise GetoptError("missing fsck type")
    fsck_type = args.pop(0)
    with Pfx(fsck_type):
      try:
        fsck_op = {
            "block": self.cmd_fsck_block,
            "dir": self.cmd_fsck_dir,
        }[fsck_type]
      except KeyError:
        raise GetoptError("unsupported fsck type")
      return fsck_op()

  def cmd_fsck_block(self, args):
    ''' Inspect a single Block.
        TODO: fromtext -> transcribe.
    '''
    xit = 0
    if not args:
      raise GetoptError("missing blockrefs")
    for blockref in args:
      with Pfx(blockref):
        blockref_bs = fromtext(blockref)
        B, offset = BlockRecord.value_from_bytes(blockref_bs)
        if offset < len(blockref_bs):
          raise ValueError("invalid blockref, extra bytes: %r" % (blockref[offset:],))
        if not fsck_Block(B):
          error("fsck failed")
          xit = 1
    return xit

  def cmd_fsck_dir(self, args):
    ''' Inspect a Dir.
    '''
    xit = 0
    if not args:
      raise GetoptError("missing dirents")
    for dirent_txt in args:
      with Pfx(dirent_txt):
        D = decode_Dirent_text(dirent_txt)
        if not fsck_dir(D):
          error("fsck failed")
          xit = 1
    return xit

  def cmd_import(self, args):
    ''' Import paths into the Store, print top Dirent for each.

        TODO: hook into vt.merge.
    '''
    xit = 0
    delete = False
    overlay = False
    whole_read = False
    opts, args = getopt(args, 'oW')
    for opt, _ in opts:
      with Pfx(opt):
        if opt == '-D':
          delete = True
        elif opt == '-o':
          overlay = True
        elif opt == '-W':
          whole_read = True
        else:
          raise RuntimeError("unhandled option: %r" % (opt,))
    if not args:
      raise GetoptError("missing path")
    srcpath = args.pop(0)
    if not args:
      raise GetoptError("missing archive.vt")
    special = args.pop(0)
    if special == '-':
      special = None
    if args:
      raise GetoptError("extra arguments: %s" % (' '.join(args),))
    if special is None:
      D = Dir('.')
    else:
      with Pfx(repr(special)):
        try:
          with open(special, 'a'):
            pass
        except OSError as e:
          error("cannot open archive for append: %s", e)
          return 1
        _, D = Archive(special).last
      if D is None:
        dstbase, suffix = splitext(basename(special))
        D = Dir(dstbase)
    with Pfx(srcpath):
      srcbase = basename(srcpath.rstrip(os.sep))
      dst = D.get(srcbase)
      if isdirpath(srcpath):
        src = OSDir(srcpath)
        if dst is None:
          dst = D.mkdir(srcbase)
        elif not dst.isdir:
          error('target name %r is not a directory', srcbase)
          xit = 1
        elif not merge(dst, src):
          error("merge failed")
          xit = 1
      elif isfilepath(srcpath):
        src = OSFile(srcpath)
        if dst is None or dst.isfile:
          D.file_fromchunks(srcbase, src.datafrom())
        else:
          error("name %r already imported: %s", srcbase, dst)
          xit = 1
      else:
        error("unsupported file type")
        xit = 1
    if special is None:
      print(D)
    else:
      with Pfx(special):
        if xit == 0:
          Archive(special).update(D)
        else:
          warning("archive not updated")
    return xit

  def cmd_init(self, args):
    ''' Install a default config and initialise the configured datadir Stores.
    '''
    if args:
      raise GetoptError("extra arguments: %r" % (args,))
    config = self.config
    config_path = config.path
    try:
      if not pathexists(config_path):
        info("write %r", config_path)
        with Pfx(config_path):
          with open(config_path, 'w') as cfg:
            self.config.write(cfg)
      basedir = config.basedir
      if not isdirpath(basedir):
        with Pfx("basedir"):
          info("mkdir %r", basedir)
          with Pfx("mkdir(%r)", basedir):
            os.mkdir(basedir)
      for clause_name, clause in sorted(config.map.items()):
        with Pfx("%s[%s]", shortpath(config_path), clause_name):
          if clause_name == 'GLOBAL':
            continue
          store_type = clause.get('type')
          if store_type == 'datadir':
            S = config[clause_name]
            S.init()
    except OSError as e:
      error("init failed: %s", e)
      return 1
    return 0

  def cmd_ls(self, args):
    ''' Do a directory listing of the specified I<dirrefs>.
    '''
    recurse = False
    if args and args[0] == "-R":
      recurse = True
      args.pop(0)
    if not args:
      raise GetoptError("missing dirrefs")
    first = True
    for path in args:
      if first:
        first = False
      else:
        print()
      D = dirent_dir(path)
      ls(path, D, recurse, sys.stdout)
    return 0

  def cmd_mount(self, args):
    ''' Mount the specified special on the specified mountpoint directory.
        Requires FUSE support.
    '''
    badopts = False
    all_dates = False
    append_only = False
    readonly = None
    opts, args = getopt(args, 'ao:r')
    for opt, val in opts:
      with Pfx(opt):
        if opt == '-a':
          all_dates = True
        elif opt == '-o':
          for option in val.split(','):
            with Pfx(option):
              if option == '':
                pass
              elif option == 'append':
                append_only = True
              elif option == 'readonly':
                readonly = True
              else:
                warning("unrecognised option")
                badopts = True
        elif opt == '-r':
          readonly = True
        else:
          raise RuntimeError("unhandled option: %r" % (opt,))
    # special is either a D{dir} or [clause] or an archive pathname
    specialD = None     # becomes not None for a D{dir}
    mount_store = defaults.S
    # the special may derive directly from a config Store clause
    special_store = None
    special_basename = None
    archive = None
    try:
      special = args.pop(0)
    except IndexError:
      special = None
      error("missing special")
      badopts = True
    else:
      with Pfx("special %r", special):
        fsname = special
        if special.startswith('D{') and special.endswith('}'):
          # D{dir}
          try:
            D, offset = parse(special)
          except ValueError as e:
            error("does not seem to be a Dir transcription: %s", e)
          else:
            if offset != len(special):
              error("unparsed text: %r", special[offset:])
            elif not isinstance(D, Dir):
              error("does not seem to be a Dir transcription, looks like a %s", type(D))
            else:
              specialD = D
          if specialD is None:
            badopts = True
          else:
            special_basename = D.name
            if not readonly:
              warning("setting readonly")
              readonly = True
        elif special.startswith('[') and special.endswith(']'):
          matched, type_, params, offset = get_store_spec(special)
          if 'clause_name' not in params:
            error("no clause name")
            badopts = True
          else:
            fsname = str(self.config) + special
            special_basename = special[1:-1].strip()
            special_store = self.config.Store_from_spec(special)
            X("special_store=%s", special_store)
            if special_store is not mount_store:
              warning(
                  "mounting using Store from special %r instead of default: %s",
                  special, mount_store)
              mount_store = special_store
            archive = self.config.archive(special_basename)
        else:
          # pathname to archive file
          archive = special
          if not isfilepath(archive):
            error("not a file: %r", archive)
            badopts = True
          else:
            fsname = shortpath(realpath(archive))
            spfx, sext = splitext(basename(special))
            if spfx and sext == '.vt':
              special_basename = spfx
            else:
              special_basename = special
    if special_basename is not None:
      # Make the name for an explicit mount safer:
      # no path components, no dots (thus no leading dots).
      special_basename = special_basename.replace(os.sep, '_').replace('.', '_')
    if args:
      mountpoint = args.pop(0)
    else:
      if special_basename is None:
        error('missing mountpoint, and cannot infer mountpoint from special: %r', special)
        badopts = True
      else:
        mountpoint = special_basename
    if args:
      subpath = args.pop(0)
    else:
      subpath = None
    if args:
      error("extra arguments: %s", ' '.join(args))
      badopts = True
    if badopts:
      raise GetoptError("bad arguments")
    if all_dates:
      readonly = True
    xit = 0
    mount_base = basename(mountpoint)
    with Pfx(special):
      if specialD is not None:
        # D{dir}
        E = specialD
      else:
        # pathname or Archive obtained from Store
        if isinstance(archive, str):
          archive = Archive(archive)
        if all_dates:
          E = Dir(mount_base)
          for when, subD in archive:
            E[datetime.fromtimestamp(when).isoformat()] = subD
        else:
          try:
            when, E = archive.last
          except OSError as e:
            error("can't access special: %s", e)
            return 1
          except ValueError as e:
            error("invalid contents: %s", e)
            return 1
          # no "last entry" (==> first use) - make an empty directory
          if E is None:
            E = Dir(mount_base)
            X("cmd_mount: new E=%r", E)
          else:
            ##dump_Dirent(E, recurse=True)
            if not E.isdir:
              error("expected directory, not file: %s", E)
              return 1
      if E.name == '.':
        info("rename %s from %r to %r", E, E.name, mount_base)
        E.name = mount_base
      # import vtfuse before doing anything with side effects
      try:
        from cs.vtfuse import mount, umount
      except ImportError as e:
        error("required module cs.vtfuse not available: %s", e)
        return 1
      with Pfx(mountpoint):
        need_rmdir = False
        if not isdirpath(mountpoint):
          # autocreate mountpoint
          info('mkdir %r ...', mountpoint)
          try:
            os.mkdir(mountpoint)
            need_rmdir = True
          except OSError as e:
            if e.errno == errno.EEXIST:
              error("mountpoint is not a directory", mountpoint)
              return 1
            raise
        T = None
        try:
          T = mount(
              mountpoint, E,
              S=mount_store, archive=archive, subpath=subpath,
              readonly=readonly, append_only=append_only,
              fsname=fsname)
          cs.x.X_via_tty = True
        except KeyboardInterrupt:
          error("keyboard interrupt, unmounting %r", mountpoint)
          xit = umount(mountpoint)
        except Exception as e:
          exception("unexpected exception: %s", e)
          xit = 1
        finally:
          if T:
            T.join()
      if need_rmdir:
        info("rmdir %r ...", mountpoint)
        try:
          os.rmdir(mountpoint)
        except OSError as e:
          error("%r: rmdir fails: %s", mountpoint, e)
          xit = 1
    return xit

  def cmd_pack(self, args):
    ''' Replace the I<path> with an archive file I<path>B<.vt> referring
        to the stored content of I<path>.
    '''
    if not args:
      raise GetoptError("missing path")
    ospath = args.pop(0)
    if args:
      raise GetoptError("extra arguments after path: %r" % (args,))
    modes = CopyModes(trust_size_mtime=True)
    with Pfx(ospath):
      if not pathexists(ospath):
        error("missing")
        return 1
      arpath = ospath + '.vt'
      A = Archive(arpath, missing_ok=True)
      when, target = A.last
      if target is None:
        target = Dir(basename(ospath))
      if isdirpath(ospath):
        source = OSDir(ospath)
      else:
        source = OSFile(ospath)
      X("target = %s, source= %s", type(target), type(source))
      if not merge(target, source):
        error("merge into %r fails", arpath)
        return 1
      A.update(target)
      info("remove %r", ospath)
      if isdirpath(ospath):
        shutil.rmtree(ospath)
      else:
        os.remove(ospath)
    return 0

  def cmd_pullfrom(self, args):
    ''' Pull missing content from other Stores.

        Usage: pullfrom other_store objects...
    '''
    if not args:
      raise GetoptError("missing other_store")
    S1spec = args.pop(0)
    if not args:
      raise GetoptError("missing objects")
    with Pfx("other_store %r", S1spec):
      S1 = Store(S1spec, self.config)
    S2 = defaults.S
    with Pfx("%s => %s", S1.name, S2.name):
      with S1:
        for obj_spec in args:
          with Pfx(obj_spec):
            try:
              obj = parse(obj_spec)
            except ValueError as e:
              raise GetoptError("unparsed: %s" % (e,)) from e
            try:
              pushto = obj.pushto
            except AttributeError:
              raise GetoptError("no pushto facility for %s objects" % (type(obj_spec),))
            pushto(S2, runstate=defaults.runstate)
    return 0

  def cmd_pushto(self, args):
    ''' Push something to a secondary Store,
        such that the secondary store has all the required Blocks.

        Usage: pushto other_store objects...
    '''
    if not args:
      raise GetoptError("missing other_store")
    S2spec = args.pop(0)
    if not args:
      raise GetoptError("missing objects")
    with Pfx("other_store %r", S2spec):
      S2 = Store(S2spec, self.config)
    S1 = defaults.S
    with Pfx("%s => %s", S1.name, S2spec):
      for obj_spec in args:
        with Pfx(obj_spec):
          try:
            obj = parse(obj_spec)
          except ValueError as e:
            raise GetoptError("unparsed: %s" % (e,)) from e
          try:
            pushto = obj.pushto
          except AttributeError:
            raise GetoptError("no pushto facility for %s objects" % (type(obj_spec),))
          pushto(S2, runstate=defaults.runstate)
    return 0

  def cmd_serve(self, args):
    ''' Start a service daemon listening on a TCP port
        or on a UNIX domain socket or on stdin/stdout.

        Usage: serve [{DEFAULT|-|/path/to/socket|[host]:port} [name:storespec]...]

        With no `name:storespec` arguments the default Store is served,
        otherwise the named Stores are exported with the first being
        served initially.
    '''
    if args:
      address = args.pop(0)
    else:
      address = 'DEFAULT'
    if address == 'DEFAULT':
      # obtain the address from the [server] config clause
      try:
        clause = self.config.get_clause('server')
      except KeyError:
        raise GetoptError(
            "no [server] clause to implement address %r"
            % (address,))
      try:
        address = clause['address']
      except KeyError:
        raise GetoptError("[server] clause: no address field")
    if not args:
      exports = {'': defaults.S}
    else:
      exports = {}
      for named_store_spec in args:
        with Pfx("name:storespec %r", named_store_spec):
          name, offset = get_identifier(named_store_spec)
          if not name:
            raise GetoptError("missing name")
          with Pfx(repr(name)):
            if name in exports:
              raise GetoptError("repeated name")
            if not named_store_spec.startswith(':', offset):
              raise GetoptError("missing colon after name")
            offset += 1
            try:
              parsed, type_, params, offset = get_store_spec(named_store_spec, offset)
            except ValueError as e:
              raise GetoptError(
                  "invalid Store specification after \"name:\": %s"
                  % (e,)) from e
            if offset < len(named_store_spec):
              raise GetoptError(
                  "extra text after storespec: %r"
                  % (named_store_spec[offset:],))
            namedS = self.config.new_Store(parsed, type_, params)
            exports[name] = namedS
            if '' not in exports:
              exports[''] = namedS
    if address == '-':
      from .stream import StreamStore
      remoteS = StreamStore(
          "serve -", sys.stdin, sys.stdout,
          exports=exports)
      remoteS.join()
    elif '/' in address:
      # path/to/socket
      socket_path = expand_path(address)
      X("serve via UNIX socket at %r", address)
      with defaults.S:
        srv = serve_socket(
            socket_path=socket_path,
            exports=exports,
            runstate=self.runstate)
      srv.join()
    else:
      # [host]:port
      cpos = address.rfind(':')
      if cpos >= 0:
        host = address[:cpos]
        port = address[cpos+1:]
        if not host:
          host = '127.0.0.1'
        port = int(port)
        with defaults.S:
          srv = serve_tcp(bind_addr=(host, port), exports=exports, runstate=self.runstate)
        srv.join()
      else:
        raise GetoptError(
            "invalid serve argument,"
            " I expect \"-\" or \"/path/to/socket\" or \"[host]:port\", got: %r"
            % (address,))
    return 0

  def cmd_test(self, args):
    ''' Test various facilites.
    '''
    if not args:
      raise GetoptError("missing test subcommand")
    subcmd = args.pop(0)
    with Pfx(subcmd):
      if subcmd == 'blockify':
        if not args:
          raise GetoptError("missing filename")
        filename = args.pop(0)
        with Pfx(filename):
          if args:
            raise GetoptError("extra arguments after filename: %r" % (args,))
          scanner = scanner_from_filename(filename)
          size_counts = defaultdict(int)
          with open(filename, 'rb') as fp:
            for chunk in blocked_chunks_of(file_data(fp, None), scanner):
              print(len(chunk), str(chunk[:16]))
              size_counts[len(chunk)] += 1
          for size, count in sorted(size_counts.items()):
            print(size, count)
        return 0
      raise GetoptError("unrecognised subcommand")

  def cmd_unpack(self, args):
    ''' Unpack the archive file _archive_`.vt` as _archive_.
    '''
    if len(args) < 1:
      raise GetoptError("missing archive name")
    arpath = args.pop(0)
    arbase, arext = splitext(arpath)
    if arext != '.vt':
      raise GetoptError("archive name does not end in .vt: %r" % (arpath,))
    if args:
      raise GetoptError("extra arguments after archive name %r" % (arpath,))
    if pathexists(arbase):
      error("archive base already exists: %r", arbase)
      return 1
    with Pfx(arpath):
      when, source = Archive(arpath).last
      if source is None:
        error("no entries in archive")
        return 1
      if source.isdir:
        target = OSDir(arbase)
      else:
        target = OSFile(arbase)
    with Pfx(arbase):
      if not merge(target, source):
        return 1
    return 0

def lsDirent(fp, E, name):
  ''' Transcribe a Dirent as an ls-style listing.
  '''
  B = E.block
  st = E.stat()
  st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size, \
      st_atime, st_mtime, st_ctime = st
  t = datetime.fromtimestamp(int(st_mtime))
  try:
    h = B.hashcode
  except AttributeError:
    detail = repr(B)
  else:
    detail = hexify(h)
  fp.write(
      "%c %-41s %s %6d %s\n"
      % (('d' if E.isdir else 'f'), detail, t, st_size, name))

def ls(path, D, recurse, fp=None):
  ''' Do an ls style directory listing with optional recursion.
  '''
  if fp is None:
    fp = sys.stdout
  fp.write('\n')
  fp.write(path)
  fp.write(":\n")
  if not recurse:
    debug("ls(): getting dirs and files...")
    names = D.dirs()+D.files()
    debug("ls(): got dirs and files = %s" % (names,))
    names.sort()
    for name in names:
      debug("ls(): D=%s (%s), name=%s" % (D, type(D), name))
      E = D[name]
      lsDirent(fp, E, name)
  else:
    dirs = D.dirs()
    dirs.sort()
    files = D.files()
    files.sort()
    for name in files:
      E = D[name]
      if E.isdir:
        warning("%s: expected file, found directory", name)
      lsDirent(fp, E, name)
    for name in dirs:
      ls(joinpath(path, name), D.chdir1(name), recurse, fp)

def cat(path, fp=None):
  ''' Write a file to the output, like cat(1).
  '''
  if fp is None:
    with os.fdopen(sys.stdout.fileno(), "wb") as bfp:
      cat(path, bfp)
  else:
    F = dirent_file(path)
    block = F.block
    for B in block.leaves:
      fp.write(B.data)

def dump(path, fp=None):
  ''' Dump the Block contents of `path`.
  '''
  if fp is None:
    fp = sys.stdout
  E, subname, unresolved = dirent_resolve(path)
  if unresolved:
    warning("%r: unresolved components: %r", path, unresolved)
  if subname:
    E = E[subname]
  dump_Block(E.block, fp)

if __name__ == '__main__':
  sys.exit(main(sys.argv))



#!/usr/bin/python

from random import randint

def rand0(maxn):
  ''' Generate a pseudorandom interger from 0 to `maxn`-1.
  '''
  return randint(0, maxn-1)

def randbool():
  ''' Return a pseudorandom Boolean value.
  '''
  return randint(0,1) == 0

def randblock(size):
  ''' Generate a pseudorandom chunk of bytes of the specified size.
  '''
  return bytes( randint(0,255) for _ in range(size) )



#!/usr/bin/env python3
#

''' Code to merge directory trees.
'''

from os.path import basename, dirname
from icontract import require
from cs.logutils import warning
from cs.pfx import Pfx
from .dir import Dir, FileDirent
from .paths import DirLike

@require(lambda target_root: isinstance(target_root, DirLike))
@require(lambda source_root: isinstance(source_root, DirLike))
def merge(target_root, source_root):
  ''' Merge contents of the DirLike `source_root`
      into the DirLike `target_root`.

      TODO: apply .stat results to merge targets.
      TODO: many modes for conflict resolution.
      TODO: whiteout entry support.
      TODO: change hash function support? or do I need 2 Stores?
  '''
  ok = True
  if not target_root.exists():
    target_root.create()
  for rpath, dirnames, filenames in source_root.walk():
    with Pfx(rpath):
      source = source_root.resolve(rpath)
      if source is None:
        warning("no longer resolves, pruning this branch")
        ok = False
        dirnames[:] = []
        filenames[:] = []
        continue
      target = target_root.resolve(rpath)
      if target is None:
        # new in target tree: mkdir the target node
        rpath_up = dirname(rpath)
        rpath_base = basename(rpath)
        target_up = target_root.resolve(rpath_up)
        target = target_up.mkdir(rpath_base)
      elif target.isdir:
        pass
      else:
        warning("conflicting item in target: not a directory")
        ok = False
      # import files
      for name in filenames:
        with Pfx(name):
          sourcef = source.get(name)
          if sourcef is None:
            # no longer available
            continue
          if sourcef.isdir:
            warning("source now a directory, skipping")
            ok = False
            continue
          targetf = target.get(name)
          if targetf is None:
            # new file
            if isinstance(target, Dir) and isinstance(sourcef, FileDirent):
              # create FileDirent from block
              target[name] = FileDirent(sourcef.block)
            else:
              # copy data
              targetf = target.file_fromchunks(name, sourcef.datafrom())
          else:
            warning("conflicting target file")
            ok = False
  return ok



#!/usr/bin/python
#
# BlockMap tests.
# - Cameron Simpson <cs@cskk.id.au>
#

''' BlockMap unit tests.
'''

import sys
from random import randint
import unittest
from .randutils import randblock
import cs.x
from cs.x import X
from .block import IndirectBlock, HashCodeBlock, _IndirectBlock
from .blockmap import BlockMap
from .store import MappingStore

cs.x.X_via_tty = True

class TestAll(unittest.TestCase):
  ''' Tests for BlockMaps.
  '''

  def setUp(self):
    ''' Unit test setup.
    '''
    self.S = MappingStore("TestAll", {})

  def _gen_data(self, depth, width):
    ''' Generate a block tree of the specified width and height filled with random data.
    '''
    if depth < 0:
      raise ValueError("depth < 0 (%s)" % (depth,))
    if width < 1:
      raise ValueError("width < 1 (%s)" % (width,))
    subblocks = []
    subchunks = []
    S = self.S
    with S:
      for _ in range(width):
        if depth == 0:
          flat_data = randblock(randint(1, 16384))
          ##flat_data = bytes(randint(1, 16384))
          top_block = HashCodeBlock(data=flat_data)
        else:
          top_block, flat_data = self._gen_data(depth-1, width)
        subblocks.append(top_block)
        subchunks.append(flat_data)
      if width == 1:
        return subblocks[0], subchunks[0]
      return IndirectBlock(subblocks=subblocks), b''.join(subchunks)

  def test001(self):
    ''' Exercise the BlockMap in various configurations.
    '''
    with self.S:
      for depth in 0, 1, 2, 3:
        for width in 1, 2, 7:   # , 17:
          X("gen depth=%d, width=%d ...", depth, width)
          top_block, flat_data = self._gen_data(depth, width)
          if not isinstance(top_block, _IndirectBlock):
            top_block = IndirectBlock(subblocks=[top_block], force=True)
          for mapsize in 999999, 10000007, 13131313, None:
            with self.subTest(mapsize=mapsize, depth=depth, width=width):
              X("test mapsize=%s", mapsize)
              bmap = BlockMap(top_block, mapsize=mapsize)
              bmap.join()
              X("full data cmp...")
              bmap.self_check()
              self.assertEqual(flat_data, bmap.data(0, len(flat_data)))
              X("full data cmp DONE")
              for _ in range(16):
                start = randint(0, len(flat_data)-1)
                end = randint(start, len(flat_data))
                X("compare [%d:%d]...", start, end)
                self.assertEqual(
                    flat_data[start:end],
                    bmap.data(start, end-start),
                    'flat_data[%d:%s] != bmap.data(%d, %d)'
                    % (start, end, start, end-start))

def selftest(argv):
  ''' Run the unit tests.
  '''
  unittest.main(__name__, None, argv)

if __name__ == '__main__':
  selftest(sys.argv)



#!/usr/bin/python
#
# Dir tests.
#       - Cameron Simpson <cs@cskk.id.au> 25aug2015
#

''' Unit tests for cs.vt.dir.
'''

from random import shuffle
import sys
import unittest
from .randutils import randbool
from .dir import FileDirent, Dir, _Dirent
from .paths import decode_Dirent_text
from .store import MappingStore

class TestAll(unittest.TestCase):
  ''' Tests for _Dirent and subclasses.
  '''

  def setUp(self):
    ''' Make a dict based Store for testing.
    '''
    self.S = MappingStore("TestAll", {})
    self.S.open()

  def tearDown(self):
    ''' Close the scratch Store.
    '''
    self.S.close()

  def _round_trip_Dirent(self, D):
    ''' Round trip the binary encode/decode.
    '''
    encoded = D.encode()
    D2, offset = _Dirent.from_bytes(encoded)
    self.assertEqual(offset, len(encoded))
    self.assertEqual(D, D2)
    text_encoded = D.textencode()
    D2 = decode_Dirent_text(text_encoded)
    self.assertEqual(D, D2)

  def test00FileDirent(self):
    ''' Trite FileDirent test.
    '''
    with self.S:
      F = FileDirent('test00')
      self._round_trip_Dirent(F)
      self.assertEqual(F.name, 'test00')

  def test01Dir(self):
    ''' Trite Dir test.
    '''
    with self.S:
      D = Dir('test01')
      self._round_trip_Dirent(D)
      self.assertEqual(D.name, 'test01')

  def test02DirRandomNames(self):
    ''' Add random entries to a Dir.
    '''
    # add random nodes
    with self.S:
      D = Dir('test02')
      self._round_trip_Dirent(D)
      dirnodes = []
      filenodes = []
      ordinals = list(range(16))
      shuffle(ordinals)
      for n in ordinals:
        dofile = randbool()
        if dofile:
          name = 'file' + str(n)
          E = FileDirent(name)
          filenodes.append(E)
        else:
          name = 'dir' + str(n)
          E = Dir(name)
          dirnodes.append(E)
        self._round_trip_Dirent(E)
        D.add(E)
        self._round_trip_Dirent(D)
        self._round_trip_Dirent(E)
        D.snapshot()
        self._round_trip_Dirent(D)
      # check that all nodes are listed as expected
      entries = dirnodes + filenodes
      shuffle(entries)
      for E in entries:
        self.assertIn(E.name, D)
        E2 = D[E.name]
        self.assertEqual(E, E2)

def selftest(argv):
  ''' Run the unit tests.
  '''
  unittest.main(__name__, None, argv)

if __name__ == '__main__':
  selftest(sys.argv)



#!/usr/bin/python
#
# Transcribe objects as text.
#   - Cameron Simpson <cs@cskk.id.au> 14jan2018
#

''' Classes to transcribe and parse textual forms of objects.

    The default syntax is slightly JSONlike syntax, though more compact and
    with less quoting noise.
    The usual form of a transcription is:

        prefix{inner}

    where `prefix` is an identifier indicating the type
    and `inner` holds the relevant details
    used to initialise the instance of the type.
    Usually `inner` takes the form:

        field:value,field:value...

    where `field` is an identifier and `value` is itself a Transcription.
    The actual form is controlled by a transcriber's `parse_inner` method.

    Some values have a more direct transcription:
    * `str`: transcribed as a quoted string
    * `int`: nonnegative integers are transcribed as bare decimal values
'''

from abc import ABC, abstractmethod
from collections import namedtuple, OrderedDict
from io import StringIO
import json
import sys
from uuid import UUID
from cs.deco import decorator
from cs.lex import get_identifier, is_identifier, \
                   get_decimal_or_float_value, get_qstr, \
                   texthexify
from cs.pfx import Pfx

class Transcriber(ABC):
  ''' Abstract base class for objects which can be used with the Transcribe class.

      Transcribers implement the following methods:
      * `transcribe_inner(T, fp)`: to transcribe to the file `fp`.
      * `parse_inner(T, s, offset, stopchar)`:
        to parse this class up to `stopchar`.

      Optional attribute:
      * `transcribe_prefix`:
        the default prefix for the "prefix{....}" markers.
  '''

  __slots__ = ()

  def __str__(self):
    return transcribe_s(self)

  @abstractmethod
  def transcribe_inner(self, T, fp):
    ''' Write the inner textual form of this object to the file `fp`.
        The result becomes "prefix{transcribe_inner()}".

        Parameters:
        * `T`: the Transcribe context
        * `fp`: the output file
    '''
    raise NotImplementedError()

  @staticmethod
  @abstractmethod
  def parse_inner(T, s, offset, stopchar, prefix):
    ''' Read the inner textual form of an object from `s` at `offset`.
        Return the object and new offset.

        Parameters:
        * `T`: the Transcribe context
        * `s`: the source text
        * `offset`: the parse position within `s`
        * `stopchar`: the end of object marker, usually '}'
        * `prefix`: the active prefix
    '''
    raise NotImplementedError()

class UUIDTranscriber:
  ''' A transcriber for uuid.UUID instances.
  '''

  @staticmethod
  def transcribe_inner(uu, fp):
    ''' Transcribe a UUID.
    '''
    fp.write(str(uu))

  @staticmethod
  def parse_inner(s, offset, stopchar, prefix):
    ''' Parse a UUID from `s` at `offset`.
        Return the UUID and the new offset.
    '''
    end_offset = s.find(stopchar, offset)
    if end_offset < offset:
      raise ValueError("offset %d: closing %r not found" % (offset, stopchar))
    uu = UUID(s[offset:end_offset])
    return uu, end_offset

ClassTranscriber = namedtuple('ClassTranscriber', 'cls transcribe_s parse')

class Transcribe:
  ''' Class to transcribe and parse textual forms of objects.
  '''

  def __init__(self):
    self.prefix_map = {}        # prefix -> baseclass
    self.class_map = {}         # baseclass -> prefix
    self.class_transcribers = {
        int: str,
        float: lambda f: "%f" % f,
        str: json.dumps,
        bool: lambda v: '1' if v else '0',
        bytes: texthexify,
        dict: lambda m: json.dumps(m, separators=(',', ':')),
        UUID: lambda u: 'U{' + str(u) + '}',
    }
    self.register(UUIDTranscriber, 'U')

  def register(self, baseclass, prefixes):
    ''' Register a class and its default prefix.

        Parameters:
        * `baseclass`: the class to register, which should be a Transcriber.
        * `prefixes`: an iterable of string prefixes to lead
          "prefix{....}" transcriptions;
          the first prefix is the default.
          This may also be a single string.
    '''
    if isinstance(prefixes, str):
      prefixes = (prefixes,)
    for prefix in prefixes:
      if prefix in self.prefix_map:
        raise ValueError("prefix %r already taken" % (prefix,))
      if ( not isinstance(baseclass, Transcriber)
           and ( not hasattr(baseclass, 'transcribe_inner')
                 or not hasattr(baseclass, 'parse_inner')
               )
         ):
        raise ValueError("baseclass %s not a subclass of Transcriber" % (baseclass,))
      self.prefix_map[prefix] = baseclass
      if baseclass not in self.class_map:
        self.class_map[baseclass] = prefix

  def register_class(self, cls, transcribe_s, parse):
    ''' Register transcribers for a class `cls`.
    '''
    class_transcribers = self.class_transcribers
    if cls in class_transcribers:
      raise ValueError("class %s already registered" % (cls,))
    class_transcribers[cls] = ClassTranscriber(cls, transcribe_s, parse)

  def transcribe(self, o, prefix=None, fp=None):
    ''' Transcribe the object `o` to file `fp`.
        `o`: the object to transcribe.
        `prefix`: prefix leading the 'prefix{...}' transcription.
          If `prefix` is None, use `o.transcribe_prefix` if defined,
          otherwise look up `type(o)` in the class_transcribers and
          use that directly.
        `fp`: optional file, default sys.stdout
    '''
    with Pfx("transcribe(%s,%r,fp)", type(o), prefix):
      if fp is None:
        fp = sys.stdout
      if prefix is None:
        # see if there is an o.transcribe_prefix
        prefix = getattr(o, 'transcribe_prefix', None)
        if prefix is None:
          # see if this class is in class_transcribers
          tr = self.class_transcribers.get(type(o))
          if tr is None:
            raise ValueError("prefix is None and no o.transcribe_prefix and no class transcriber")
          fp.write(tr(o))
          return
      # use prefix and rely on o.transcribe_inner
      baseclass = self.prefix_map.get(prefix)
      if baseclass is not None:
        if not isinstance(o, baseclass):
          raise ValueError("type(o)=%s, not an instanceof(%r)" % (type(o), baseclass))
      fp.write(prefix)
      fp.write('{')
      o.transcribe_inner(self, fp)
      fp.write('}')

  def transcribe_s(self, o, prefix=None):
    ''' Transcribe the object `o` to a string.
        `o`: the object to transcribe.
        `prefix`: optional marker prefix
    '''
    fp = StringIO()
    self.transcribe(o, prefix=prefix, fp=fp)
    s = fp.getvalue()
    fp.close()
    return s

  def transcribe_mapping(self, m, fp):
    ''' Transcribe the mapping `m` to the file `fp`.
        `m`: the mapping to transcribe.
        `fp`: optional file, default sys.stdout
        The keys of the mapping must be identifiers.
        Values which are None are skipped.
    '''
    with Pfx("transcribe_mapping(%r)", m):
      first = True
      for k, v in m.items():
        if not is_identifier(k):
          raise ValueError("not an identifier key: %r" % (k,))
        if v is None:
          continue
        if not first:
          fp.write(',')
        fp.write(k)
        fp.write(':')
        self.transcribe(v, None, fp)
        first = False

  def parse(self, s, offset=0):
    ''' Parse an object from the string `s` starting at `offset`.
        Return the object and the new offset.

        Parameters:
        * `s`: the source string
        * `offset`: optional string offset, default 0
    '''
    ##X("parse %r ...", s[offset:])
    # strings
    value, offset2 = self.parse_qs(s, offset, optional=True)
    if value is not None:
      return value, offset2
    # decimal values
    if s[offset:offset+1].isdigit():
      return get_decimal_or_float_value(s, offset)
    # {json}
    if s.startswith('{', offset):
      sub = s[offset:]
      m, suboffset = json.JSONDecoder().raw_decode(sub)
      offset += suboffset
      return m, offset
    # prefix{....}
    prefix, offset = get_identifier(s, offset)
    if not prefix:
      raise ValueError("no type prefix at offset %d" % (offset,))
    with Pfx("prefix %r", prefix):
      if offset >= len(s) or s[offset] != '{':
        raise ValueError("missing opening '{' at offset %d" % (offset,))
      offset += 1
      baseclass = self.prefix_map.get(prefix)
      if baseclass is None:
        raise ValueError("prefix not registered: %r" % (prefix,))
      o, offset = baseclass.parse_inner(self, s, offset, '}', prefix)
      if offset > len(s):
        raise ValueError("parse_inner returns offset beyond text")
      if offset >= len(s) or s[offset] != '}':
        raise ValueError("missing closing '}' at offset %d" % (offset,))
      offset += 1
      return o, offset

  @staticmethod
  def parse_qs(s, offset=0, optional=False):
    ''' Parse a quoted string from `s` at `offset`.
        Return the string value and the new offset.

        Parameters:
        * `s`: the source string
        * `offset`: optional string offset, default 0
        * `optional`: if true (default False), return None if there
          is no quoted string at offset instead of raising a ValueError
    '''
    if s.startswith("'", offset) or s.startswith('"', offset):
      return get_qstr(s, offset=offset, q=s[offset])
    if optional:
      return None, offset
    raise ValueError("offset %d: expected quoted string" % (offset,))

  def parse_mapping(
      self,
      s, offset=0, stopchar=None,
      required=None, optional=None
  ):
    ''' Parse a mapping from the string `s`.
        Return the mapping and the new offset.

        Parameters:
        * `s`: the source string
        * `offset`: optional string offset, default 0
        * `stopchar`: ending character, not to be consumed
        * `required`: if specified, validate that the mapping contains
          all the keys in this list
        * `optional`: if specified, validate that the mapping contains
          no keys which are not required or optional

        If `required` or `optional` is specified the return takes the form:

            offset, required_values..., optional_values...

        where missing optional values are presented as None.
    '''
    if optional is not None and required is None:
      raise ValueError("required is None but optional is specified: %r" % (optional,))
    d = OrderedDict()
    while offset < len(s) and (stopchar is None or s[offset] != stopchar):
      k, offset = get_identifier(s, offset)
      if not k:
        raise ValueError("offset %d: not an identifier" % (offset,))
      if offset >= len(s) or s[offset] != ':':
        raise ValueError("offset %d: expected ':'" % (offset,))
      offset += 1
      v, offset = self.parse(s, offset)
      d[k] = v
      if offset >= len(s):
        break
      c = s[offset]
      if c == stopchar:
        break
      if c != ',':
        raise ValueError("offset %d: expected ',' but found: %r" % (offset, s[offset:]))
      offset += 1
    if required is None and optional is None:
      return d, offset
    for k in required:
      if k not in d:
        raise ValueError("missing required field %r" % (k,))
    if optional is not None:
      for k in d.keys():
        if k not in required and k not in optional:
          raise ValueError("unexpected field %r" % (k,))
    ret = [offset]
    for k in required:
      ret.append(d[k])
    for k in optional:
      ret.append(d.get(k))
    return ret

# global default Transcribe context
_TRANSCRIBE = Transcribe()

def register(cls, prefix=None, T=None):
  ''' Register a class and prefix with a Transcribe context.

      Parameters:
      * `cls`: the class to register
      * `prefix`: the marker prefix. If not specified, use `cls.transcribe_prefix`.
      * `T`: the transcribe context, default: `_TRANSCRIBE`
  '''
  global _TRANSCRIBE
  if prefix is None:
    prefix = cls.transcribe_prefix
  if T is None:
    T = _TRANSCRIBE
  return T.register(cls, prefix)

def transcribe(o, prefix=None, fp=None, T=None):
  ''' Transcribe the object `o` to file `fp`.

      Parameters:
      * `o`: the object to transcribe.
      * `prefix`: optional marker prefix
      * `fp`: optional file, default sys.stdout
      * `T`: the transcribe context, default: `_TRANSCRIBE`
  '''
  global _TRANSCRIBE
  if fp is None:
    fp = sys.stdout
  if T is None:
    T = _TRANSCRIBE
  return T.transcribe(o, prefix, fp)

def transcribe_s(o, prefix=None, T=None):
  ''' Transcribe the object `o` to a string.

      Parameters:
      * `o`: the object to transcribe.
      * `prefix`: optional marker prefix
      * `T`: the transcribe context, default: `_TRANSCRIBE`
  '''
  global _TRANSCRIBE
  if T is None:
    T = _TRANSCRIBE
  return T.transcribe_s(o, prefix)

def transcribe_mapping(m, fp, T=None):
  ''' Transcribe the mapping `m` to the file `fp`.

      Parameters:
      * `m`: the mapping to transcribe.
      * `fp`: optional file, default sys.stdout
      * `T`: the transcribe context, default: `_TRANSCRIBE`

      The keys of the mapping must be identifiers.
  '''
  global _TRANSCRIBE
  if T is None:
    T = _TRANSCRIBE
  return T.transcribe_mapping(m, fp)

def parse(s, offset=0, T=None):
  ''' Parse an object from the string `s`. Return the object and the new offset.

      Parameters:
      * `s`: the source string
      * `offset`: optional string offset, default 0
      * `T`: the transcribe context, default: `_TRANSCRIBE`
  '''
  global _TRANSCRIBE
  if T is None:
    T = _TRANSCRIBE
  return T.parse(s, offset)

def parse_mapping(
    s, offset=0, stopchar=None, T=None,
    required=None, optional=None
):
  ''' Parse a mapping from the string `s`.
      Return the mapping and the new offset.

      Parameters:
      * `s`: the source string
      * `offset`: optional string offset, default 0
      * `stopchar`: ending character, not to be consumed
      * `T`: the transcribe context, default: `_TRANSCRIBE`
  '''
  global _TRANSCRIBE
  if T is None:
    T = _TRANSCRIBE
  return T.parse_mapping(
      s, offset, stopchar,
      required=required, optional=optional)

@decorator
def mapping_transcriber(
    cls, *, prefix=None, T=None,
    transcription_mapping=None,
    required=None, optional=None,
    factory=None,
):
  ''' A class decorator to provide mapping style `parse_inner` and
      `transcribe_inner` methods and to register the class against
      a Transcribe instance.

      Parameters:
      * `prefix`: the prefix string
      * `T`: the transcribe instance, default: `_TRANSCRIBE`
      * `transcription_mapping`: a function of `self` to produce
        the mapping to be transcribed
      * `required`: optional list of keys required in the mapping
      * `optional`: optional list of keys which may be present in
        the mapping
      * `factory`: a factory to construct an instance of `cls` given
        keywords arguments supplied by the parsed mapping;
        default: `cls`

      Example:

          @mapping_transcriber(
              prefix="Ino",
              transcription_mapping=lambda self: {
                  'refcount': self.recount,
                  'E': self.E,
              },
              required=('refcount', 'E'),
              optional=(),
          )
          class Inode(Transcriber):
  '''
  if prefix is None:
    raise ValueError("missing prefix")
  if transcription_mapping is None:
    raise ValueError(
        "missing transcription_mapping, expected a function of self")
  if factory is None:
    factory = cls
  @classmethod
  def parse_inner(cls, T, s, offset, stopchar, parsed_prefix):
    ''' Parse the inner section as a mapping.
    '''
    if parsed_prefix != prefix:
      raise ValueError(
          "expected prefix=%r, got: %r"
          % (prefix, parsed_prefix,))
    m, offset = parse_mapping(
        s, offset, stopchar=stopchar, T=T,
        required=required, optional=optional)
    return factory(**m), offset
  cls.parse_inner = parse_inner
  def transcribe_inner(self, T, fp):
    return transcribe_mapping(transcription_mapping(self), fp, T=T)
  cls.transcribe_inner = transcribe_inner
  register(cls, prefix=prefix, T=T)
  return cls

if __name__ == '__main__':
  from .transcribe_tests import selftest
  selftest()



#!/usr/bin/env python3
#
# Filesystem layer for Dirs.
# This is used by systems like FUSE.
# - Cameron Simpson <cs@cskk.id.au>

''' Filesystem semantics for a Dir.
'''

import errno
import os
from os import O_CREAT, O_RDONLY, O_WRONLY, O_RDWR, O_APPEND, O_TRUNC, O_EXCL, O_NOFOLLOW
import shlex
from types import SimpleNamespace as NS
from uuid import UUID
from cs.excutils import logexc
from cs.later import Later
from cs.logutils import exception, error, warning, info, debug
from cs.pfx import Pfx
from cs.range import Range
from cs.threads import locked
from cs.x import X
from . import defaults, Lock, RLock
from .block import isBlock
from .cache import BlockCache
from .dir import _Dirent, Dir, FileDirent
from .debug import dump_Dirent
from .meta import Meta
from .parsers import scanner_from_filename, scanner_from_mime_type
from .paths import resolve
from .transcribe import Transcriber, mapping_transcriber, parse

XATTR_VT_PREFIX = 'x-vt-'

DEFAULT_FS_THREAD_MAX = 16

def oserror(errno_, msg, *a):
  ''' Function to issue a warning and then raise an OSError.
  '''
  assert isinstance(errno_, int)
  assert isinstance(msg, str)
  if a:
    msg = msg % a
  warning("raise OSError(%s): %s", errno_, msg)
  raise OSError(errno_, msg)

OS_EEXIST = lambda msg, *a: oserror(errno.EEXIST, msg, *a)
OS_EFAULT = lambda msg, *a: oserror(errno.EFAULT, msg, *a)
OS_EINVAL = lambda msg, *a: oserror(errno.EINVAL, msg, *a)
OS_ELOOP = lambda msg, *a: oserror(errno.ELOOP, msg, *a)
OS_ENOATTR = lambda msg, *a: oserror(errno.ENOATTR, msg, *a)
OS_ENOENT = lambda msg, *a: oserror(errno.ENOENT, msg, *a)
OS_ENOTDIR = lambda msg, *a: oserror(errno.ENOTDIR, msg, *a)
OS_ENOTSUP = lambda msg, *a: oserror(errno.ENOTSUP, msg, *a)
OS_EROFS = lambda msg, *a: oserror(errno.EROFS, msg, *a)

class FileHandle:
  ''' Filesystem state for an open file.
  '''

  def __init__(self, fs, E, for_read, for_write, for_append, lock=None):
    ''' Initialise the FileHandle with filesystem, dirent and modes.
    '''
    if lock is None:
      lock = Lock()
    self.fs = fs
    self.E = E
    self.for_read = for_read
    self.for_write = for_write
    self.for_append = for_append
    self._block_mapping = None
    self._lock = lock
    E.open()

  def __str__(self):
    fhndx = getattr(self, 'fhndx', None)
    return "<FileHandle:fhndx=%d:%s>" % (fhndx, self.E,)

  def bg(self, func, *a, **kw):
    ''' Function dispatcher.
    '''
    return self.fs.bg(func, *a, **kw)

  def close(self):
    ''' Close the file, mark its parent directory as changed.
    '''
    S = defaults.S
    R = self.E.flush()
    self.E.parent.changed = True
    S.open()
    # NB: additional S.open/close around self.E.close
    R.notify(logexc(lambda _: (
        defaults.pushStore(S),
        self.E.close(),
        defaults.popStore(),
        S.close()
    )))

  def write(self, data, offset):
    ''' Write data to the file.
    '''
    f = self.E.open_file
    with f:
      with self._lock:
        if self.for_append and offset != len(f):
          OS_EFAULT(
              "%s: file open for append but offset(%s) != length(%s)",
              f, offset, len(f))
        f.seek(offset)
        written = f.write(data)
    self.E.touch()
    return written

  def read(self, size, offset):
    ''' Read data from the file.
    '''
    if size < 1:
      raise ValueError("FileHandle.read: size(%d) < 1" % (size,))
    return self.E.open_file.read(size, offset=offset, longread=True)

  def truncate(self, length):
    ''' Truncate the file, mark it as modified.
    '''
    self.E.open_file.truncate(length)
    self.E.touch()

  def flush(self):
    ''' Commit file contents to Store.
        Chooses a scanner based on the Dirent.name.
    '''
    X("FileHandle.flush: self.E.name=%r", self.E.name)
    mime_type = self.E.meta.mime_type
    if mime_type is None:
      scanner = None
    else:
      X("look up scanner from mime_type %r", mime_type)
      scanner = scanner_from_mime_type(mime_type)
    if scanner is None:
      X("look up scanner from filename %r", self.E.name)
      scanner = scanner_from_filename(self.E.name)
    self.E.flush(scanner, dispatch=self.bg)
    ## no touch, already done by any writes
    X("FileHandle.Flush DONE")

@mapping_transcriber(
    prefix="Ino",
    transcription_mapping=lambda self: {
        'refcount': self.refcount,
        'E': self.E,
    },
    required=('refcount', 'E'),
    optional=(),
)
class Inode(Transcriber, NS):
  ''' An Inode associates an inode number and a Dirent.

      Attributes:
      * `inum`: the inode number
      * `E`: the primary Dirent
      * `refcount`: the number of Dir references to this Dirent
  '''

  transcribe_prefix = 'Ino'

  def __init__(self, inum, E, refcount=1):
    NS.__init__(self)
    self.inum = inum
    self.E = E
    self.refcount = refcount

  def __repr__(self):
    return (
        "%s(inum=%d,refcount=%d,E=%s(%r))"
        % (
            type(self).__name__,
            self.inum,
            self.refcount,
            type(self.E).__name__, self.E.name
        )
    )

  def transcribe_inner(self, T, fp):
    return T.transcribe_mapping({
        'refcount': self.refcount,
        'E': self.E,
    }, fp, T=T)

  @classmethod
  def parse_inner(cls, T, s, offset, stopchar, prefix):
    if prefix != cls.transcribe_prefix:
      raise ValueError(
          "expected prefix=%r, got: %r"
          % (cls.transcribe_prefix, prefix,))
    m, offset = T.parse_mapping(s, offset, stopchar=stopchar, T=T)
    return cls(None, m['E'], m['refcount']), offset

class Inodes(object):
  ''' Inode information for a filesystem.

      This consists of:
      - a Range denoting allocated inode numbers
      - a mapping of inode numbers to Inodes
      - a mapping of UUIDs to Inodes
      - a mapping of Dirents to Inodes

      Once an Inode is allocated it will have a reference by inum
      and Dirent. Since a Dirent need not have a UUID, it may not
      be mapped by UUID. The UUID map will be updated if `.add` is
      called later when the Dirent has a UUID, and clients should
      call `.add` to ensure that mapping if they rely on a Dirent's
      UUID, such as when making an IndirectDirent.
  '''

  def __init__(self, fs):
    self.fs = fs                # main filesystem
    self._allocated = Range()   # range of allocated inode numbers
    self._by_inum = {}
    self._by_uuid = {}
    self._by_dirent = {}
    self._lock = RLock()

  def load_fs_inode_dirents(self, D):
    ''' Load entries from an `fs_inode_dirents` Dir into the Inode table.
    '''
    X("LOAD FS INODE DIRENTS:")
    dump_Dirent(D)
    for name, E in D.entries.items():
      X("  name=%r, E=%r", name, E)
      with Pfx(name):
        # get the refcount from the :uuid:refcount" name
        _, refcount_s = name.split(':')[:2]
        I = self.add(E)
        I.refcount = int(refcount_s)
        X("  I=%s", I)

  def get_fs_inode_dirents(self):
    ''' Create an `fs_inode_dirents` Dir containing Inodes which
        should be preserved across mounts.
    '''
    D = Dir('fs_inode_dirents')
    for uuid, I in sorted(self._by_uuid.items()):
      if I.refcount > 0:
        D["%s:%d" % (uuid, I.refcount)] = I.E
      else:
        warning("refcount=%s, SKIP %s", I.refcount, I.E)
    X("GET FS INODE DIRENTS:")
    dump_Dirent(D)
    return D

  def _new_inum(self):
    ''' Allocate a new Inode number.
    '''
    allocated = self._allocated
    if allocated:
      span0 = allocated.span0
      inum = span0.end
    else:
      inum = 1
    allocated.add(inum)
    return inum

  def add(self, E, inum=None):
    ''' Add the Dirent `E` to the Inodes, return the new Inode.
        It is not an error to add the same Dirent more than once.
    '''
    with Pfx("Inodes.add(E=%s)", E):
      if E.isindirect:
        raise ValueError("indirect Dirents may not become Inodes")
      if inum is not None and inum < 1:
        raise ValueError("inum must be >= 1, got: %d" % (inum,))
      uu = E.uuid
      I = self._by_dirent.get(E)
      if I:
        assert I.E is E
        if inum is not None and I.inum != inum:
          raise ValueError(
              "inum=%d: Dirent already has an Inode with a different inum: %s"
              % (inum, I))
        if uu:
          # opportunisticly update UUID mapping
          # in case the Dirent has acquired a UUID
          I2 = self._by_uuid.get(uu)
          if I2:
            assert I2.E is E
          else:
            self._by_uuid[uu] = I
        return I
      # unknown Dirent, create new Inode
      if inum is None:
        inum = self._new_inum()
      else:
        I = self._by_inum.get(inum)
        if I:
          raise ValueError("inum %d already allocated: %s" % (inum, I))
        self._allocated.add(inum)
      I = Inode(inum, E)
      self._by_dirent[E] = I
      self._by_inum[inum] = I
      if uu:
        self._by_uuid[uu] = I
      return I

  def __getitem__(self, ndx):
    if isinstance(ndx, int):
      try:
        I = self._by_inum[ndx]
      except KeyError as e:
        raise IndexError("unknown inode number %d: %s" % (ndx, e))
      return I
    if isinstance(ndx, UUID):
      return self._by_uuid[ndx]
    if isinstance(ndx, _Dirent):
      return self._by_dirent[ndx]
    raise TypeError("cannot deference indices of type %r" % (type(ndx),))

  def __contains__(self, ndx):
    try:
      _ = self[ndx]
    except (KeyError, IndexError):
      return False
    return True

class FileSystem(object):
  ''' The core filesystem functionality supporting FUSE operations
      and in principle other filesystem-like access.

      See the cs.vtfuse module for the StoreFS_LLFUSE class (aliased
      as StoreFS) and associated mount function which presents a
      FileSystem as a FUSE mount.

      TODO: medium term: see if this can be made into a VFS layer
      to support non-FUSE operation, for example a VT FTP client
      or the like.
  '''

  def __init__(
      self, E,
      *,
      S=None,
      archive=None,
      subpath=None,
      readonly=None,
      append_only=False,
      show_prev_dirent=False,
      thread_max=None,
  ):
    ''' Initialise a new mountpoint.

        Parameters:
        * `E`: the root directory reference
        * `S`: the backing Store
        * `archive`: if not None, an Archive or similar, with a
          `.update(Dirent[,when])` method
        * `subpath`: relative path to mount Dir
        * `readonly`: forbid data modification
        * `append_only`: append only mode: files may only grow,
          filenames may not be changed or deleted
        * `show_prev_dirent`: show previous Dir revision as the '...' entry
    '''
    if not E.isdir:
      raise ValueError("not dir Dir: %s" % (E,))
    if S is None:
      S = defaults.S
    self._old_S_block_cache = S.block_cache
    self.block_cache = S.block_cache or defaults.block_cache or BlockCache()
    S.block_cache = self.block_cache
    S.open()
    if readonly is None:
      readonly = S.readonly
    if thread_max is None:
      thread_max = DEFAULT_FS_THREAD_MAX
    self.E = E
    self.S = S
    self.archive = archive
    if archive is None:
      self._last_sync_state = None
    else:
      self._last_sync_state = bytes(E)
    self.subpath = subpath
    self.readonly = readonly
    self.append_only = append_only
    self.show_prev_dirent = show_prev_dirent
    if subpath:
      # locate subdirectory to display at mountpoint
      mntE, _, tail_path = resolve(E, subpath)
      if tail_path:
        raise ValueError("subpath %r does not resolve" % (subpath,))
      if not mntE.isdir:
        raise ValueError("subpath %r is not a directory" % (subpath,))
      self.mntE = mntE
    else:
      mntE = E
    self.mntE = mntE
    self.device_id = -1
    self._fs_uid = os.geteuid()
    self._fs_gid = os.getegid()
    self._lock = RLock()
    self._later = Later(DEFAULT_FS_THREAD_MAX)
    self._path_files = {}
    self._file_handles = []
    inodes = self._inodes = Inodes(self)
    self[1] = mntE
    try:
      with Pfx("fs_inode_dirents"):
        fs_inode_dirents = E.meta.get("fs_inode_dirents")
        X("FS INIT: fs_inode_dirents=%s", fs_inode_dirents)
        if fs_inode_dirents:
          inode_dir, offset = _Dirent.from_str(fs_inode_dirents)
          if offset < len(fs_inode_dirents):
            warning("unparsed text after Dirent: %r", fs_inode_dirents[offset:])
          X("IMPORT INODES:")
          dump_Dirent(inode_dir)
          inodes.load_fs_inode_dirents(inode_dir)
        else:
          X("NO INODE IMPORT")
        X("FileSystem mntE:")
      with self.S:
        with defaults.stack('fs', self):
          dump_Dirent(mntE)
    except Exception as e:
      exception("exception during initial report: %s", e)

  def bg(self, func, *a, **kw):
    ''' Dispatch a function via the FileSystem's Later instance.
    '''
    return self._later.defer(func, *a, **kw)

  def close(self):
    ''' Close the FileSystem.
    '''
    self._sync()
    self.S.close()
    self.S.block_cache = self._old_S_block_cache

  def __str__(self):
    if self.subpath:
      return "<%s S=%s /=%s %r=%s>" % (
          self.__class__.__name__,
          self.S, self.E, self.subpath, self.mntE
      )
    return "%s(S=%s,/=%r)" % (type(self).__name__, self.S, self.E)

  def __getitem__(self, inum):
    ''' Lookup inode numbers or UUIDs.
    '''
    return self._inodes[inum]

  def __setitem__(self, inum, E):
    ''' Associate a specific inode number with a Dirent.
    '''
    self._inodes.add(E, inum)

  @logexc
  def _sync(self):
    with Pfx("_sync"):
      if defaults.S is None:
        raise RuntimeError("RUNTIME: defaults.S is None!")
      archive = self.archive
      if not self.readonly and archive is not None:
        with self._lock:
          E = self.E
          updated = False
          X("snapshot %r  ...", E)
          E.snapshot()
          X("snapshot: afterwards E=%r", E)
          fs_inode_dirents = self._inodes.get_fs_inode_dirents()
          X("_SYNC: FS_INODE_DIRENTS:")
          dump_Dirent(fs_inode_dirents)
          X("set meta.fs_inode_dirents")
          if fs_inode_dirents.size > 0:
            E.meta['fs_inode_dirents'] = str(fs_inode_dirents)
          else:
            E.meta['fs_inode_dirents'] = ''
          new_state = bytes(E)
          if new_state != self._last_sync_state:
            archive.update(E)
            self._last_sync_state = new_state
            updated = True
        # debugging
        if updated:
          dump_Dirent(E, recurse=False)

  def _resolve(self, path):
    ''' Call paths.resolve and return its result.
    '''
    return resolve(self.mntE, path)

  def _namei2(self, path):
    ''' Look up path. Raise OSError(ENOENT) if missing. Return Dirent, parent.
    '''
    E, P, tail_path = self._resolve(path)
    if tail_path:
      OS_ENOENT("cannot resolve path %r", path)
    return E, P

  def _namei(self, path):
    ''' Look up path. Raise OSError(ENOENT) if missing. Return Dirent.
    '''
    E, _ = self._namei2(path)
    return E

  @locked
  def E2inode(self, E):
    ''' Return the Inode for the supplied Dirent `E`.
    '''
    if E.isindirect:
      E = E.ref
    return self._inodes.add(E)

  def i2E(self, inum):
    ''' Return the Dirent associated with the supplied `inum`.
    '''
    I = self._inodes[inum]
    X("inum %s => %r", inum, I)
    return I.E

  def open2(self, P, name, flags):
    ''' Open a regular file given parent Dir `P` and `name`,
        allocate FileHandle, return FileHandle index.

        Increments the kernel reference count.
        Wraps self.open.
    '''
    if not P.isdir:
      OS_ENOTDIR("parent (name=%r) not a directory", P.name)
    if name in P:
      if flags & O_EXCL:
        OS_EEXIST("entry %r already exists", name)
      E = P[name]
    elif not flags & O_CREAT:
      OS_ENOENT("no entry named %r", name)
    else:
      E = FileDirent(name)
      P[name] = E
    return self.open(E, flags)

  def open(self, E, flags):
    ''' Open a regular file `E`, allocate FileHandle, return FileHandle index.
        Increments the kernel reference count.
    '''
    for_read = (flags & O_RDONLY) == O_RDONLY or (flags & O_RDWR) == O_RDWR
    for_write = (flags & O_WRONLY) == O_WRONLY or (flags & O_RDWR) == O_RDWR
    for_append = (flags & O_APPEND) == O_APPEND
    for_trunc = (flags & O_TRUNC) == O_TRUNC
    debug("for_read=%s, for_write=%s, for_append=%s",
          for_read, for_write, for_append)
    if for_trunc and not for_write:
      OS_EINVAL("O_TRUNC requires O_WRONLY or O_RDWR")
    if for_append and not for_write:
      OS_EINVAL("O_APPEND requires O_WRONLY or O_RDWR")
    if (for_write and not for_append) and self.append_only:
      OS_EINVAL("fs is append_only but no O_APPEND")
    if for_trunc and self.append_only:
      OS_EINVAL("fs is append_only but O_TRUNC")
    if (for_write or for_append) and self.readonly:
      error("fs is readonly")
      OS_EROFS("fs is readonly")
    if E.issym:
      if flags & O_NOFOLLOW:
        OS_ELOOP("open symlink with O_NOFOLLOW")
      OS_EINVAL("open(%s)" % (E,))
    elif not E.isfile:
      OS_EINVAL("open of nonfile: %s" % (E,))
    FH = FileHandle(self, E, for_read, for_write, for_append, lock=self._lock)
    if flags & O_TRUNC:
      FH.truncate(0)
    return self._new_file_handle_index(FH)

  @locked
  def _fh(self, fhndx):
    try:
      fh = self._file_handles[fhndx]
    except IndexError:
      error("cannot look up FileHandle index %r", fhndx)
      raise
    return fh

  def _fh_remove(self, fhndx):
    self._file_handles[fhndx] = None

  def _fh_close(self, fhndx):
    fh = self._fh(fhndx)
    fh.close()
    self._fh_remove(fhndx)

  @locked
  def _new_file_handle_index(self, file_handle):
    ''' Allocate a new FileHandle index for a `file_handle`.

        TODO: linear allocation cost, may need recode if things get
          busy; might just need a list of released fds for reuse.
    '''
    fhs = self._file_handles
    for fhndx, fh in enumerate(fhs):
      if fh is None:
        fhs[fhndx] = file_handle
        return fhndx
    fhs.append(file_handle)
    return len(fhs) - 1

  @staticmethod
  def access(E, amode, uid=None, gid=None):
    ''' Check access mode `amode` against Dirent `E`.
    '''
    with Pfx("access(E=%r,amode=%s,uid=%r,gid=%d)", E, amode, uid, gid):
      # test the access against the caller's uid/gid
      # pass same in as default file ownership in case there are no metadata
      return E.meta.access(
          amode, uid, gid, default_uid=uid, default_gid=gid)

  def getxattr(self, inum, xattr_name):
    ''' Get the extended attribute `xattr_name` from `inum`.
    '''
    E = self.i2E(inum)
    xattr_name = Meta.xattrify(xattr_name)
    if xattr_name.startswith(XATTR_VT_PREFIX):
      # process special attribute names
      suffix = xattr_name[len(XATTR_VT_PREFIX):]
      if suffix == 'block':
        return str(E.block).encode()
      OS_EINVAL(
          "getxattr(inum=%s,xattr_name=%r): invalid %r prefixed name",
          inum, xattr_name, XATTR_VT_PREFIX)
    xattr = E.meta.getxattr(xattr_name, None)
    if xattr is None:
      ##if xattr_name == 'com.apple.FinderInfo':
      ##  OS_ENOTSUP("inum %d: no xattr %r, pretend not supported", inum, xattr_name)
      OS_ENOATTR("inum %d: no xattr %r", inum, xattr_name)
    return xattr

  def removexattr(self, inum, xattr_name):
    ''' Remove the extended attribute named `xattr_name` from `inum`.
    '''
    if self.readonly:
      OS_EROFS("fs is read only")
    E = self.i2E(inum)
    xattr_name = Meta.xattrify(xattr_name)
    if xattr_name.startswith(XATTR_VT_PREFIX):
      OS_EINVAL(
          "removexattr(inum=%s,xattr_name=%r): invalid %r prefixed name",
          inum, xattr_name, XATTR_VT_PREFIX)
    meta = E.meta
    try:
      meta.delxattr(xattr_name)
    except KeyError:
      OS_ENOATTR("no such extended attribute: %r", xattr_name)

  def setxattr(self, inum, xattr_name, xattr_value):
    ''' Set the extended attribute `xattr_name` to `xattr_value`
        on inode `inum`.
    '''
    if self.readonly:
      OS_EROFS("fs is read only")
    E = self.i2E(inum)
    xattr_name = Meta.xattrify(xattr_name)
    if not xattr_name.startswith(XATTR_VT_PREFIX):
      # ordinary attribute, set it and return
      E.meta.setxattr(xattr_name, xattr_value)
      return
    # process special attribute names
    with Pfx("%s.setxattr(%d,%r,%r)", self, inum, xattr_name, xattr_value):
      suffix = xattr_name[len(XATTR_VT_PREFIX):]
      with Pfx(suffix):
        if suffix == 'block':
          # update the Dirent's content directly
          if not E.isfile:
            OS_EINVAL("tried to update the data content of a nonfile: %s", E)
          block_s = Meta.xattrify(xattr_value)
          B, offset = parse(block_s)
          if offset < len(block_s):
            OS_EINVAL(
                "unparsed text after trancription: %r",
                block_s[offset:])
          if not isBlock(B):
            OS_EINVAL("not a Block transcription")
          info("%s: update .block directly to %r", E, str(B))
          E.block = B
          return
        if suffix == 'control':
          argv = shlex.split(xattr_value.decode('utf-8'))
          if not argv:
            OS_EINVAL("no control command")
          op = argv.pop(0)
          with Pfx(op):
            if op == 'cache':
              if argv:
                OS_EINVAL("extra arguments: %r", argv)
              B = E.block
              if B.indirect:
                X("ADD BLOCK CACHE FOR %s", B)
                bm = self.block_cache.get_blockmap(B)
                X("==> BLOCKMAP: %s", bm)
              else:
                X("IGNORE BLOCK CACHE for %s: not indirect", B)
              return
            OS_EINVAL("unrecognised control command")
        OS_EINVAL("invalid %r prefixed name", XATTR_VT_PREFIX)



#!/usr/bin/python
#
# Store tests.
#       - Cameron Simpson <cs@cskk.id.au>
#

import os
from os.path import abspath
import random
import shutil
import sys
import tempfile
import unittest
from cs.excutils import logexc
from cs.logutils import setup_logging, warning
from .randutils import randblock
from cs.x import X
import cs.x
cs.x.X_via_tty = True
from . import _TestAdditionsMixin
from .index import class_names as get_index_names, class_by_name as get_index_by_name
from .hash import HashUtilDict, DEFAULT_HASHCLASS, HASHCLASS_BY_NAME
from .hash_tests import _TestHashCodeUtils
from .store import MappingStore, DataDirStore, ProgressStore

def multitest(method):
  ''' Decorator to permute a test method for multplie Store types and hash classes.
  '''
  def testMethod(self):
    setups = [
        ('MappingStore', MappingStore, {'mapping': {}}),
    ]
    for index_name in get_index_names():
      setups.append(
          ('DataDirStore', DataDirStore,
            {'indexclass': get_index_by_name(index_name),
             'rollover': 200000})
      )
    for test_store, factory, kwargs, *args in setups:
      for hashclass_name in sorted(HASHCLASS_BY_NAME.keys()):
        hashclass = HASHCLASS_BY_NAME[hashclass_name]
        with self.subTest(test_store=test_store, hashclass=hashclass_name):
          args = ["%s:%s:%s" % (test_store, hashclass_name, method.__name__)] + args
          T = self.mktmpdir(prefix=method.__module__+'.'+method.__name__)
          with T as tmpdirpath:
            if factory in (DataDirStore,):
              args.insert(1, tmpdirpath)
            self.S = factory(*args, hashclass=hashclass, **kwargs)
            with self.S:
              method(self)
              self.S.flush()
  return testMethod

class TestStore(unittest.TestCase, _TestAdditionsMixin):

  @multitest
  def test00empty(self):
    S = self.S
    self.assertLen(S, 0)

  @multitest
  def test01add_new_block(self):
    S = self.S
    self.assertLen(S, 0)
    size = random.randint(127, 16384)
    data = randblock(size)
    # compute block hash but do not store
    h = S.hash(data)
    self.assertLen(S, 0)
    ok = S.contains(h)
    self.assertFalse(ok)
    self.assertNotIn(h, S)
    # now add the block
    h2 = S.add(data)
    self.assertEqual(h, h2)
    self.assertLen(S, 1)
    ok = S.contains(h)
    self.assertTrue(ok)
    self.assertIn(h, S)

  @multitest
  def test02add_get(self):
    S = self.S
    self.assertLen(S, 0)
    random_chunk_map = {}
    for _ in range(16):
      size = random.randint(127, 16384)
      data = randblock(size)
      h = S.hash(data)
      h2 = S.add(data)
      self.assertEqual(h, h2)
      random_chunk_map[h] = data
    self.assertLen(S, 16)
    for h in random_chunk_map:
      chunk = S.get(h)
      self.assertIsNot(chunk, None)
      self.assertEqual(chunk, random_chunk_map[h])

##class TestMappingStore(TestStore, unittest.TestCase):
##
##  def _init_Store(self):
##    self.S = MappingStore("TestMappingStore", {}, hashclass=self.hashclass)
##
##class TestProgressStore(TestStore, unittest.TestCase):
##
##  def _init_Store(self):
##    M = MappingStore("TestProgressStore", {}, hashclass=self.hashclass)
##    MO = M.open()
##    P = ProgressStore("ProgressMappingStore", MO)
##    self.S = P.open()

class TestHashCodeUtilsMappingStoreDict(_TestHashCodeUtils, unittest.TestCase):
  ''' Test HashUtils on a MappingStore on a plain dict.
  '''
  MAP_FACTORY = lambda self: MappingStore("TestHashCodeUtilsMappingStoreDict", {}, hashclass=DEFAULT_HASHCLASS)

class TestHashCodeUtilsMappingStoreHashUtilDict(_TestHashCodeUtils, unittest.TestCase):
  ''' Test HashUtils on a MappingStore on a HashUtilDict.
  '''
  MAP_FACTORY = lambda self: MappingStore("TestHashCodeUtilsMappingStoreHashUtilDict", HashUtilDict(), hashclass=DEFAULT_HASHCLASS)

try:
  import kyotocabinet
except ImportError:
  pass
else:
  class TestDataDirStoreKyoto(_TestDataDirStore, unittest.TestCase):
    INDEX_CLASS = KyotoIndex
  class TestHashCodeUtilsDataDirStoreKyotoStore(_TestHashCodeUtils, unittest.TestCase):
    MAP_FACTORY = lambda self: DataDirStore("TestHashCodeUtilsDataDirStoreKyotoStore", self.mktmpdir(), hashclass=DEFAULT_HASHCLASS, indexclass=KyotoIndex, rollover=200000)

def selftest(argv):
  unittest.main(__name__, None, argv)

if __name__ == '__main__':
  setup_logging(__file__)
  selftest(sys.argv)



#!/usr/bin/python

''' A content hash based data store with a filesystem layer, using
    variable sized blocks, arbitrarily sized data and utilising some
    domain knowledge to aid efficient block boundary selection.

    See also the Plan 9 Venti system:
    (http://library.pantek.com/general/plan9.documents/venti/venti.html,
    http://en.wikipedia.org/wiki/Venti).
'''

from collections import namedtuple
import os
from string import ascii_letters, digits
import tempfile
import threading
from threading import (
    Lock as threading_Lock,
    RLock as threading_RLock,
    current_thread,
)
from cs.lex import texthexify, untexthexify
from cs.logutils import error, warning
from cs.mappings import StackableValues
from cs.py.stack import caller, stack_dump
from cs.seq import isordered
import cs.resources
from cs.resources import RunState
from cs.x import X

DISTINFO = {
    'keywords': ["python3"],
    'classifiers': [
        ##"Development Status :: 3 - Alpha",
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
    ],
    'install_requires': [
        'cs.buffer',
        'cs.app.flag',
        'cs.binary',
        'cs.binary_tests',
        'cs.cache',
        'cs.debug',
        'cs.deco',
        'cs.env',
        'cs.excutils',
        'cs.fileutils',
        'cs.inttypes',
        'cs.later',
        'cs.lex',
        'cs.logutils',
        'cs.mappings',
        'cs.packetstream',
        'cs.pfx',
        'cs.progress',
        'cs.py.func',
        'cs.py.stack',
        'cs.queues',
        'cs.range',
        'cs.resources',
        'cs.result',
        'cs.seq',
        'cs.serialise',
        'cs.socketutils',
        'cs.threads',
        'cs.tty',
        'cs.units',
        'cs.x',
        'lmdb',
    ],
    'entry_points': {
        'console_scripts': [
            'vt = cs.vt.__main__:main',
        ],
    },
}

DEFAULT_CONFIG_PATH = '~/.vtrc'

DEFAULT_BASEDIR = '~/.vt_stores'

DEFAULT_CONFIG = {
    'GLOBAL': {
      'basedir': DEFAULT_BASEDIR,
      'blockmapdir': '[default]/blockmaps',
    },
    'default': {
      'type': 'datadir',
      'path': 'trove',
    },
    'cache': {
      'type': 'memory',
      'max_data': '16 GiB',
    },
    'server': {
      'type': 'datadir',
      'path': 'trove',
      'address': '~/.vt.sock',
    },
}

# intercept Lock and RLock
if False:
  def RLock():
    ''' Obtain a recursive DebuggingLock.
    '''
    return DebuggingLock(recursive=True)
  def Lock():
    ''' Obtain a nonrecursive DebuggingLock.
    '''
    return DebuggingLock()
  # monkey patch MultiOpenMixin
  cs.resources._mom_lockclass = RLock
else:
  Lock = threading_Lock
  RLock = threading_RLock

# Default OS level file high water mark.
# This is used for rollover levels for DataDir files and cache files.
MAX_FILE_SIZE = 1024 * 1024 * 1024

# path separator, hardwired
PATHSEP = '/'

class _Defaults(threading.local, StackableValues):
  ''' Per-thread default context stack.

      A Store's __enter__/__exit__ methods push/pop that store
      from the `.S` attribute.
  '''

  _Ss = []  # global stack of fallback Store values

  def __init__(self):
    threading.local.__init__(self)
    StackableValues.__init__(self)
    self.push('runstate', RunState())
    self.push('fs', None)
    self.push('block_cache', None)

  def _fallback(self, key):
    ''' Fallback function for empty stack.
    '''
    if key == 'S':
      warning("no per-Thread Store stack, using the global stack")
      stack_dump()
      Ss = self._Ss
      if Ss:
        return Ss[-1]
      error("%s: no per-Thread defaults.S and no global stack, returning None", self)
      return None
    raise ValueError("no fallback for %r" % (key,))

  def pushStore(self, newS):
    ''' Push a new Store onto the per-Thread stack.
    '''
    self.push('S', newS)

  def popStore(self):
    ''' Pop and return the topmost Store from the per-Thread stack.
    '''
    oldS = self.pop('S')
    return oldS

  def push_Ss(self, newS):
    ''' Push a new Store onto the global stack.
    '''
    self._Ss.append(newS)

  def pop_Ss(self):
    ''' Pop and return the topmost Store from the global stack.
    '''
    return self._Ss.pop()

  def push_runstate(self, new_runstate):
    ''' Context manager to push a new RunState instance onto the per-Thread stack.
    '''
    return self.stack('runstate', new_runstate)

defaults = _Defaults()

def fromtext(s):
  ''' Return raw byte array from text/hexadecimal string.
  '''
  return untexthexify(s)

# Characters that may appear in text sections of a texthexify result.
# Because we transcribe Dir blocks this way it includes some common
# characters used for metadata, notably including the double quote
# because it is heavily using in JSON.
# It does NOT include '/' because these appear at the start of paths.
_TEXTHEXIFY_WHITE_CHARS = ascii_letters + digits + '_+-.,=:;{"}*'

def totext(data):
  ''' Represent a byte sequence as a hex/text string.
  '''
  return texthexify(data, whitelist=_TEXTHEXIFY_WHITE_CHARS)

class _TestAdditionsMixin:
  ''' Some common methods uses in tests.
  '''

  @classmethod
  def mktmpdir(cls, prefix=None):
    ''' Create a temporary directory.
    '''
    if prefix is None:
      prefix = cls.__qualname__
    return tempfile.TemporaryDirectory(
        prefix="test-" + prefix + "-",
        suffix=".tmpdir",
        dir=os.getcwd()
    )

  def assertLen(self, o, length, *a, **kw):
    ''' Test len(o) unless it raises TypeError.
    '''
    try:
      olen = len(o)
    except TypeError:
      pass
    else:
      self.assertEqual(olen, length, *a, **kw)

  def assertIsOrdered(self, s, reverse, strict=False):
    ''' Assertion to test that an object's elements are ordered.
    '''
    self.assertTrue(
        isordered(s, reverse, strict),
        "not ordered(reverse=%s,strict=%s): %r" % (reverse, strict, s))

LockContext = namedtuple("LockContext", "caller thread")

class DebuggingLock(object):
  ''' A wrapper for a threading Lock or RLock
      to notice contention and report contending uses.
  '''

  def __init__(self, recursive=False):
    self.recursive = recursive
    self.trace_acquire = False
    self._lock = threading_RLock() if recursive else threading_Lock()
    self._held = None

  def __repr__(self):
    return "%s(lock=%r,held=%s)" % (type(self).__name__, self._lock, self._held)

  def acquire(self, timeout=-1, _caller=None):
    ''' Acquire the lock and note the caller who takes it.
    '''
    if _caller is None:
      _caller = caller()
    lock = self._lock
    hold = LockContext(_caller, current_thread())
    if timeout != -1:
      warning(
          "%s:%d: lock %s: timeout=%s",
          hold.caller.filename, hold.caller.lineno,
          lock, timeout)
    contended = False
    if True:
      if lock.acquire(0):
        lock.release()
      else:
        contended = True
        held = self._held
        warning(
            "%s:%d: lock %s: waiting for contended lock, held by %s:%s:%d",
            hold.caller.filename, hold.caller.lineno,
            lock,
            held.thread, held.caller.filename, held.caller.lineno)
    acquired = lock.acquire(timeout=timeout)
    if contended:
      warning(
          "%s:%d: lock %s: %s",
          hold.caller.filename, hold.caller.lineno,
          lock,
          "acquired" if acquired else "timed out")
    self._held = hold
    if acquired and self.trace_acquire:
      X("ACQUIRED %r", self)
      stack_dump()
    return acquired

  def release(self):
    ''' Release the lock and forget who took it.
    '''
    self._held = None
    self._lock.release()

  def __enter__(self):
    ##X("%s.ENTER...", type(self).__name__)
    self.acquire(_caller=caller())
    ##X("%s.ENTER: acquired, returning self", type(self).__name__)
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    self.release()
    return False

  def _is_owned(self):
    lock = self._lock
    return lock._is_owned()



#!/usr/bin/python
#
# File interfaces.      - Cameron Simpson <cs@cskk.id.au>
#

''' Classes to present blocks as file-like objects with write support.
'''

from __future__ import print_function, absolute_import
from io import RawIOBase
from os import SEEK_SET
import sys
from cs.fileutils import BackedFile, ReadMixin, datafrom
from cs.resources import MultiOpenMixin
from cs.result import bg
from cs.threads import locked, LockableMixin
from . import defaults, RLock
from .block import Block, IndirectBlock, RLEBlock
from .blockify import top_block_for, blockify

# arbitrary threshold to generate blockmaps
AUTO_BLOCKMAP_THRESHOLD = 1024 * 1024

class ROBlockFile(RawIOBase, ReadMixin):
  ''' A read-only file interface to a Block based on io.RawIOBase.
  '''

  def __init__(self, block):
    ''' Initialise with Block `block`.
    '''
    RawIOBase.__init__(self)
    self.isdir = False
    self.block = block
    self._offset = 0

  def __len__(self):
    ''' Length of the file, as the length of the backing Block.
    '''
    return len(self.block)

  def seek(self, offset, whence=0):
    ''' Set the current file offset.
    '''
    if whence == 1:
      offset += self.tell()
    elif whence == 2:
      offset += len(self)
    self._offset = offset
    return offset

  def tell(self):
    ''' Return the current file offset.
    '''
    return self._offset

  def datafrom(self, offset):
    ''' Generator yielding natural chunks from the file commencing at offset.
        This supports the ReadMixin.read method.
    '''
    # data from the backing block
    backing_block = self.block
    if len(backing_block) >= AUTO_BLOCKMAP_THRESHOLD:
      backing_block.get_blockmap()
    return backing_block.datafrom(offset)

class RWBlockFile(MultiOpenMixin, LockableMixin, ReadMixin):
  ''' A read/write file-like object based on cs.fileutils.BackedFile.

      An initial Block is supplied for use as the backing data.
      The .flush and .close methods return a new Block representing the commited data.

      *Note*: a RWBlockFile starts open and must be closed.
  '''

  def __init__(self, backing_block=None):
    ''' Initialise RWBlockFile with optional backing Block `backing_block`.
    '''
    if backing_block is None:
      backing_block = Block(data=b'')
    self.filename = None
    self._syncer = None     # syncing Result, close waits for it
    self._backing_block = None
    self._blockmap = None
    self._file = None
    self._lock = RLock()
    MultiOpenMixin.__init__(self, lock=self._lock)
    self.open()
    self._reset(backing_block)

  def __str__(self):
    return "RWBlockFile(backing_block=%s)" % (self._backing_block,)

  def _reset(self, new_backing_block):
    ''' Discard the file contents and replace with the supplied Block, unmodified.
    '''
    old_backing_block = self._backing_block
    if old_backing_block is not new_backing_block:
      if old_backing_block is not None:
        try:
          old_backing_block.blockmap = None
        except AttributeError:
          pass
      if self._file is not None:
        self._file.close()
      self._backing_block = new_backing_block
      if new_backing_block is None:
        self._file = None
      else:
        self._file = BackedFile(ROBlockFile(new_backing_block))
        self._file.flush = self.flush
        self._sync_span = None

  def startup(self):
    ''' Startup actions.
    '''
    pass

  def shutdown(self):
    ''' Close the RWBlockFile, return the top Block.
    '''
    B = self.sync()
    self._reset(None)
    return B

  def __len__(self):
    f = self._file
    if f is None:
      length = len(self._backing_block)
    else:
      length = len(f)
      span = self._sync_span
      if span is not None:
        length = max(length, span.end)
    return length

  @property
  @locked
  def backing_block(self):
    ''' Return the current backing block.

        Note: the backing block may be out of date with respect to any
        pending flushes; call .sync() to obtain an up to date flushed
        block.
    '''
    return self._backing_block

  @locked
  def flush(self, scanner=None, dispatch=None):
    ''' Push the current state to the Store and update the current top block.
        Return a Result which completes later.

        We dispatch the sync in the background within a lock.

        Parameters:
        * `scanner`: optional scanner for new file data to locate
          preferred block boundaries.
    '''
    if dispatch is None:
      dispatch = bg
    syncer = self._syncer
    if syncer is None:
      S = defaults.S
      S.open()
      syncer = self._syncer = dispatch(self._sync_file, S, scanner=scanner)
      def cleanup(R):
        with self._lock:
          if R is self._syncer:
            self._syncer = None
      syncer.notify(cleanup)
    return syncer

  def sync(self):
    ''' Dispatch a flush, return the flushed backing block.
        Wait for any flush to complete before returing the backing block.
    '''
    return self.flush()()

  def _sync_file(self, S, scanner=None):
    # worker to sync the front ranges to the Block store
    f = self._file
    while f.front_range:
      span = None
      with self._lock:
        if f.front_range:
          span = f.front_range._spans.pop(0)
      if span is None:
        break
      start, end = span
      with S:
        self._sync_span = span
        new_block = file_top_block(f.front_file, start, end, scanner=scanner)
        old_backing_block = self._backing_block
        if start >= len(old_backing_block):
          # old_block + pad + new_block
          subblocks = [old_backing_block]
          pad_length = start - len(old_backing_block)
          if pad_length > 0:
            subblocks.append(RLEBlock(pad_length, b'\0'))
          subblocks.append(new_block)
          new_backing_block = IndirectBlock(subblocks=subblocks)
        else:
          end = min(end, len(old_backing_block))
          new_backing_block = old_backing_block.splice(start, end, new_block)
      # update the backing file, leave the front file alone
      new_file = ROBlockFile(new_backing_block)
      with self._lock:
        self._backing_block = new_backing_block
        f.back_file = new_file
        self._sync_span = None
    S.close()
    return self._backing_block

  @locked
  def truncate(self, length):
    ''' Truncate the RWBlockFile to the specified `length`.
    '''
    if length < 0:
      raise ValueError("length must be >= 0, received %s" % (length,))
    if length == 0:
      self._reset(Block(data=b''))
    else:
      # let any syncers complete
      self.sync()
      cur_len = len(self)
      f = self._file
      front_range = f.front_range
      front_file = f.front_file
      backing_block0 = self.backing_block
      if length < cur_len:
        # shorten file
        if front_range.end > length:
          front_range.discard_span(cur_len, front_range.end)
          # the front_file should also be too big
          front_file.truncate(length)
        if len(backing_block0) > length:
          # new top Block built on previous Block
          # this might overlap some of the front_range but the only new blocks
          # should be the partial direct block at the end of the range, and
          # whatever new indirect blocks get made to span things
          self.backing_block \
              = top_block_for(backing_block0.top_blocks(0, length))
      elif length > cur_len:
        # extend the front_file and front_range
        front_file.truncate(length)
        front_range.add_span(front_range.end, length)

  def tell(self):
    ''' Return the file read/write position.
    '''
    return self._file.tell()

  def seek(self, offset, whence=SEEK_SET):
    ''' Adjust the file read/write position.
    '''
    return self._file.seek(offset, whence=whence)

  def write(self, data):
    ''' Write `data` to the RWBlockFile.
    '''
    return self._file.write(data)

  def datafrom(self, offset):
    ''' Generator yielding natural chunks from the file commencing at offset.
        This supports the ReadMixin.read method.
    '''
    ##raise RuntimeError("BANG")
    f = self._file
    front_file = f.front_file
    backing_block = self.backing_block
    if len(backing_block) >= AUTO_BLOCKMAP_THRESHOLD:
      backing_block.get_blockmap()
    # TODO: iterate afresh from each offset? expensive, but accomodates changes
    for inside, (start, end) in f.front_range.slices(offset, len(self)):
      if inside:
        # data from the front file
        yield from filedata(front_file, start, end)
      else:
        # sync span data are also obtained from the front file
        sync_span = self._sync_span
        if sync_span is not None:
          sync_start, sync_end = sync_span
          if start < sync_start < end:
            # preamble from the backing block: start:sync_start
            yield from backing_block.datafrom(start=start, end=sync_start)
            start = sync_start
          if start < sync_end and end > sync_start:
            # overlap from front file: start:min(end,sync_end)
            new_start = min(end, sync_end)
            yield from filedata(front_file, start, new_start)
            start = new_start
          if start == end:
            # no post sync span data
            continue
          assert start < end
        # data from the backing block
        for bs in backing_block.datafrom(start=start, end=end):
          yield bs

def filedata(f, start, end):
  ''' A generator to yield chunks of data from a file.
      These chunks don't need to be preferred-edge aligned;
      blockify() does that.
  '''
  return datafrom(f, start, maxlength=end-start)

def file_top_block(f, start, end, scanner=None):
  ''' Return a top Block for the data from an open file.
  '''
  return top_block_for(blockify(filedata(f, start, end), scanner=scanner))

if __name__ == '__main__':
  from .file_tests import selftest
  selftest(sys.argv)
